##
# Copyright 2013-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing Qt, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""
import os
from easybuild.tools import LooseVersion

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import apply_regex_substitutions
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.systemtools import get_cpu_architecture, get_glibc_version, get_shared_lib_ext
from easybuild.tools.systemtools import AARCH64, POWER, RISCV64


class EB_Qt(ConfigureMake):
    """
    Support for building and installing Qt.
    """

    @staticmethod
    def extra_options():
        extra_vars = {
            'check_qtwebengine': [False, "Make sure QtWebEngine components is installed", CUSTOM],
            'disable_advanced_kernel_features': [False, "Disable features that require a kernel > 3.15", CUSTOM],
            'platform': [None, "Target platform to build for (e.g. linux-g++-64, linux-icc-64)", CUSTOM],
        }
        extra_vars = ConfigureMake.extra_options(extra_vars)

        # allowing to specify prefix_opt doesn't make sense for Qt, since -prefix is hardcoded in configure_step
        del extra_vars['prefix_opt']

        return extra_vars

    def configure_step(self):
        """Configure Qt using interactive `configure` script."""

        self.cfg.update('configopts', '-release')

        platform = None
        comp_fam = self.toolchain.comp_family()
        if self.cfg['platform']:
            platform = self.cfg['platform']
        # if no platform is specified, try to derive it based on compiler in toolchain
        elif comp_fam in [toolchain.GCC]:  # @UndefinedVariable
            myarch = get_cpu_architecture()
            if myarch in [AARCH64, RISCV64]:
                platform = 'linux-g++'
            else:
                platform = 'linux-g++-64'
        elif comp_fam in [toolchain.INTELCOMP]:  # @UndefinedVariable
            if LooseVersion(self.version) >= LooseVersion('4'):
                platform = 'linux-icc-64'
            else:
                platform = 'linux-icc'
                # fix -fPIC flag (-KPIC is not correct for recent Intel compilers)
                qmake_conf = os.path.join('mkspecs', platform, 'qmake.conf')
                apply_regex_substitutions(qmake_conf, [('-KPIC', '-fPIC')])

        if platform:
            self.cfg.update('configopts', "-platform %s" % platform)
        else:
            raise EasyBuildError("Don't know which platform to set based on compiler family.")

        if LooseVersion(self.version) >= LooseVersion('5.8'):
            # Qt5 doesn't respect $CFLAGS, $CXXFLAGS and $LDFLAGS, but has equivalent compiler options,
            # e.g. QMAKE_CFLAGS; see https://doc.qt.io/qt-5/qmake-variable-reference.html#qmake-cc.
            # Since EasyBuild relies e.g. for --optarch on $CFLAGS, we need to
            # set the equivalent QMAKE_* configure options.
            # (see also https://github.com/easybuilders/easybuild-easyblocks/issues/1670)
            env_to_options = {
                'CC': 'QMAKE_CC',
                'CFLAGS': 'QMAKE_CFLAGS',
                'CXX': 'QMAKE_CXX',
                'CXXFLAGS': 'QMAKE_CXXFLAGS',
                # QMAKE_LFLAGS is not a typo, see: https://doc.qt.io/qt-5/qmake-variable-reference.html#qmake-lflags
                'LDFLAGS': 'QMAKE_LFLAGS',
            }
            for env_name, option in sorted(env_to_options.items()):
                value = os.getenv(env_name)
                if value is not None:
                    if env_name.endswith('FLAGS'):
                        # For *FLAGS, we add to existing flags (e.g. those set in Qt's .pro-files).
                        config_opt = option + '+="%s"'
                    else:
                        # For compilers, we replace QMAKE_CC/CXX
                        # (otherwise, you get e.g. QMAKE_CC="g++ g++", which fails)
                        config_opt = option + '="%s"'

                    self.cfg.update('configopts', config_opt % value)

        # configure Qt such that xmlpatterns is also installed
        # -xmlpatterns is not a known configure option for Qt 5.x, but there xmlpatterns support is enabled by default
        if LooseVersion(self.version) >= LooseVersion('4') and LooseVersion(self.version) < LooseVersion('5'):
            self.cfg.update('configopts', '-xmlpatterns')

        # disable specific features to avoid that libQt5Core.so being tagged as requiring kernel 3.17,
        # which causes confusing problems like this even though the file exists and can be found by...
        #     error while loading shared libraries: libQt5Core.so.5:
        #      cannot open shared object file: No such file or directory
        # see also:
        # * https://bugs.gentoo.org/669994
        # * https://github.com/NixOS/nixpkgs/commit/a7b6a9199e8db54a798d011a0946cdeb72cfc46b
        # * https://gitweb.gentoo.org/proj/qt.git/commit/?id=9ff0752e1ee3c28818197eaaca45545708035152
        kernel_version = os.uname()[2]
        skip_kernel_features = self.cfg['disable_advanced_kernel_features']
        old_kernel_version = LooseVersion(kernel_version) < LooseVersion('3.17')
        if LooseVersion(self.version) >= LooseVersion('5.10') and (skip_kernel_features or old_kernel_version):
            self.cfg.update('configopts', '-no-feature-renameat2')
            self.cfg.update('configopts', '-no-feature-getentropy')

        cmd = "%s ./configure -prefix %s %s" % (self.cfg['preconfigopts'], self.installdir, self.cfg['configopts'])
        qa = [
            (r"Type 'o' if you want to use the Open Source Edition\.", 'o'),
            (r"Do you accept the terms of either license\?", 'yes'),
            (r"Which edition of Qt do you want to use\?", 'o'),
        ]
        no_qa = [
            "for .*pro",
            r"%s.*" % os.getenv('CXX', '').replace('+', '\\+'),  # need to escape + in 'g++'
            "Reading .*",
            "WARNING .*",
            "Project MESSAGE:.*",
            "rm -f .*",
            'Creating qmake...',
            'Checking for .*...',
        ]
        run_shell_cmd(cmd, qa_patterns=qa, qa_wait_patterns=no_qa, qa_timeout=120)

        # Ninja uses all visible cores by default, which can lead to lack of sufficient memory;
        # so $NINJAFLAGS is set to control number of parallel processes used by Ninja;
        # note that $NINJAFLAGS is not a generic thing for Ninja, it's very specific to the Qt5 build procedure
        if LooseVersion(self.version) >= LooseVersion('5'):
            if get_software_root('Ninja'):
                env.setvar('NINJAFLAGS', self.parallel_flag)

    def build_step(self):
        """Set $LD_LIBRARY_PATH before calling make, to ensure that all required libraries are found during linking."""
        # cfr. https://elist.ornl.gov/pipermail/visit-developers/2011-September/010063.html

        if LooseVersion(self.version) >= LooseVersion('5.6'):
            libdirs = ['qtbase', 'qtdeclarative']
        else:
            libdirs = ['']

        libdirs = [os.path.join(self.cfg['start_dir'], d, 'lib') for d in libdirs]
        self.cfg.update('prebuildopts', 'LD_LIBRARY_PATH=%s' % os.pathsep.join(libdirs + ['$LD_LIBRARY_PATH']))

        super().build_step()

    def sanity_check_step(self):
        """Custom sanity check for Qt."""

        shlib_ext = get_shared_lib_ext()

        if LooseVersion(self.version) >= LooseVersion('4'):
            libversion = ''
            if LooseVersion(self.version) >= LooseVersion('5'):
                libversion = self.version.split('.')[0]

            libfile = os.path.join('lib', 'libQt%sCore.%s' % (libversion, shlib_ext))

        else:
            libfile = os.path.join('lib', 'libqt.%s' % shlib_ext)

        custom_paths = {
            'files': ['bin/moc', 'bin/qmake', libfile],
            'dirs': ['include', 'plugins'],
        }

        if self.cfg['check_qtwebengine']:
            glibc_version = get_glibc_version()
            myarch = get_cpu_architecture()
            if LooseVersion(glibc_version) <= LooseVersion("2.16"):
                self.log.debug("Skipping check for qtwebengine, since it requires a more recent glibc.")
            elif myarch == POWER:
                self.log.debug("Skipping check for qtwebengine, since it is not supported on POWER.")
            else:
                qtwebengine_libs = ['libQt%s%s.%s' % (libversion, x, shlib_ext) for x in ['WebEngine', 'WebEngineCore']]
                custom_paths['files'].extend([os.path.join('lib', lib) for lib in qtwebengine_libs])

        if LooseVersion(self.version) >= LooseVersion('4'):
            custom_paths['files'].append('bin/xmlpatterns')

        super().sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
Geant4 support, implemented as an easyblock.

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""

import os

from easybuild.easyblocks.generic.cmakemake import CMakeMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools import LooseVersion


class EB_Geant4(CMakeMake):
    """
    Support for building Geant4.
    """

    @staticmethod
    def extra_options():
        """
        Define extra options needed by Geant4
        """
        extra_vars = CMakeMake.extra_options()
        extra_vars.update({
            'G4ABLAVersion': [None, "G4ABLA version", CUSTOM],
            'G4NDLVersion': [None, "G4NDL version", CUSTOM],
            'G4EMLOWVersion': [None, "G4EMLOW version", CUSTOM],
            'PhotonEvaporationVersion': [None, "PhotonEvaporation version", CUSTOM],
            'G4RadioactiveDecayVersion': [None, "G4RadioactiveDecay version", CUSTOM],
        })
        # Requires out-of-source build
        extra_vars['separate_build_dir'][0] = True
        return extra_vars

    def make_module_extra(self):
        """Define Geant4-specific environment variables in module file."""
        g4version = '.'.join(self.version.split('.')[:2])

        # '10.01.p03' -> '10.1.3'
        shortver = self.version.replace('.0', '.').replace('.p0', '.')
        datadst = os.path.join(self.installdir, 'share', '%s-%s' % (self.name, shortver), 'data')

        txt = super().make_module_extra()
        txt += self.module_generator.set_environment('G4INSTALL', self.installdir)
        # no longer needed in > 9.5, but leave it there for now.
        txt += self.module_generator.set_environment('G4VERSION', g4version)

        incdir = os.path.join(self.installdir, 'include')
        txt += self.module_generator.set_environment('G4INCLUDE', os.path.join(incdir, 'Geant4'))
        txt += self.module_generator.set_environment('G4LIB', os.path.join(self.installdir, 'lib64', 'Geant4'))

        if self.cfg['PhotonEvaporationVersion']:
            g4levelgammadata = os.path.join(datadst, 'PhotonEvaporation%s' % self.cfg['PhotonEvaporationVersion'])
            txt += self.module_generator.set_environment('G4LEVELGAMMADATA', g4levelgammadata)

        if self.cfg['G4RadioactiveDecayVersion']:
            g4radioactivedata = os.path.join(datadst, 'RadioactiveDecay%s' % self.cfg['G4RadioactiveDecayVersion'])
            txt += self.module_generator.set_environment('G4RADIOACTIVEDATA', g4radioactivedata)

        if self.cfg['G4EMLOWVersion']:
            g4ledata = os.path.join(datadst, 'G4EMLOW%s' % self.cfg['G4EMLOWVersion'])
            txt += self.module_generator.set_environment('G4LEDATA', g4ledata)

        if self.cfg['G4NDLVersion']:
            g4neutronhpdata = os.path.join(datadst, 'G4NDL%s' % self.cfg['G4NDLVersion'])
            txt += self.module_generator.set_environment('G4NEUTRONHPDATA', g4neutronhpdata)

        return txt

    def sanity_check_step(self):
        """
        Custom sanity check for Geant4
        """
        bin_files = ["bin/geant4-config", "bin/geant4.sh", "bin/geant4.csh"]
        libs = ['analysis', 'event', 'GMocren', 'materials', 'readout', 'Tree', 'VRML']

        # G4Persistency library was split up in Geant v11.2,
        # see https://geant4.web.cern.ch/download/release-notes/notes-v11.2.0.html
        if LooseVersion(self.version) >= LooseVersion('11.2'):
            libs.extend(['gdml', 'geomtext', 'mctruth', 'geomtext'])
        else:
            libs.append('persistency')

        lib_files = ["lib64/libG4%s.so" % x for x in libs]
        include_dir = 'include/Geant4'

        custom_paths = {
            'files': bin_files + lib_files,
            'dirs': [include_dir],
        }

        super().sanity_check_step(custom_paths)



##
# Copyright 2015-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://vscentrum.be/nl/en),
# the Hercules foundation (http://www.herculesstichting.be/in_English)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing Paraver, implemented as an easyblock

@author Kenneth Hoste (Ghent University)
@author Markus Geimer (Juelich Supercomputing Centre)
@author Bernd Mohr (Juelich Supercomputing Centre)
"""
import glob
import os
from easybuild.tools import LooseVersion

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_libdir, get_software_root


class EB_Paraver(ConfigureMake):
    """Support for building/installing Paraver."""

    def __init__(self, *args, **kwargs):
        """Constructor for custom easyblock for Paraver."""
        super().__init__(*args, **kwargs)

        if LooseVersion(self.version) < LooseVersion('4.7'):
            raise EasyBuildError("Custom easyblock for Paraver only supports Paraver versions >= 4.7")

    def configure_step(self):
        """Custom configuration procedure for Paraver: template configuration options before using them."""

        # first, check (required) dependencies
        # this is done in configure_step to avoid breaking 'installation' with --module-only --force
        boost_root = get_software_root('Boost')
        if not boost_root:
            raise EasyBuildError("Boost is not available as a dependency")

        wx_config = None
        wxwidgets = get_software_root('wxWidgets')
        if wxwidgets:
            wx_config = os.path.join(wxwidgets, 'bin', 'wx-config')
        else:
            raise EasyBuildError("wxWidgets is not available as a dependency")

        # determine value to pass to --with-wxpropgrid (library name)
        wxpropgrid = None
        wxpropertygrid_root = get_software_root('wxPropertyGrid')
        if wxpropertygrid_root:
            wxpropertygrid_libdir = os.path.join(wxpropertygrid_root, get_software_libdir('wxPropertyGrid'))
            libname_pattern = 'libwxcode_*_propgrid-*.so'
            wxpropgrid_libs = glob.glob(os.path.join(wxpropertygrid_libdir, libname_pattern))
            if len(wxpropgrid_libs) == 1:
                # exclude the 'lib' prefix and '.so' suffix to determine value to be passed to --with-wxpropgrid
                wxpropgrid = os.path.basename(wxpropgrid_libs[0])[3:-3]
            else:
                tup = (libname_pattern, wxpropertygrid_libdir, wxpropgrid_libs)
                raise EasyBuildError("Expected to find exactly one %s library in %s, found %s" % tup)
        else:
            self.log.info("wxPropertyGrid not included as dependency, assuming that's OK...")

        self.cfg.update('configopts', '--with-boost=%s' % boost_root)
        self.cfg.update('configopts', '--with-paraver=%s' % self.installdir)
        self.cfg.update('configopts', '--with-wx-config=%s' % wx_config)
        # wxPropertyGrid is not required with recent wxWidgets
        if wxpropgrid:
            self.cfg.update('configopts', '--with-wxpropgrid=%s' % wxpropgrid)

        super().configure_step()

    def build_step(self):
        """No build ('make') required for recent versions."""
        pass

    def sanity_check_step(self):
        """Custom sanity check for Paraver."""
        custom_paths = {
            'files': ['bin/wxparaver', 'include/paraverconfig.h', 'lib/paraver-kernel/libparaver-kernel.a'],
            'dirs': [],
        }
        super().sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing the Intel Fortran compiler suite, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Ward Poelmans (Ghent University)
"""

import os

from easybuild.easyblocks.generic.intelbase import IntelBase
from easybuild.easyblocks.icc import EB_icc  # @UnresolvedImport
from easybuild.tools import LooseVersion
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import MODULE_LOAD_ENV_HEADERS
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_ifort(EB_icc, IntelBase):
    """
    Class that can be used to install ifort
    - minimum version suported: 2020.0
    - will fail for all older versions (due to newer silent installer)
    """

    def __init__(self, *args, **kwargs):
        """Constructor, initialize class variables."""
        super().__init__(*args, **kwargs)

        if LooseVersion(self.version) < LooseVersion('2020'):
            raise EasyBuildError(
                f"Version {self.version} of {self.name} is unsupported. Mininum supported version is 2020.0."
            )

        # define list of subdirectories in search paths of module load environment
        # add additional paths to those of ICC only needed for separate ifort installations
        for envar in self.module_load_environment.alias(MODULE_LOAD_ENV_HEADERS):
            envar.append(os.path.join(self.comp_libs_subdir, 'compiler/include'))

    def sanity_check_step(self):
        """Custom sanity check paths for ifort."""
        shlib_ext = get_shared_lib_ext()

        binprefix = 'bin'
        binfiles = ['ifort']
        binaries = [os.path.join(binprefix, f) for f in binfiles]

        libprefix = 'lib/intel64'
        libfiles = [f'lib{lib}' for lib in ['ifcore.a', f'ifcore.{shlib_ext}', 'iomp5.a', f'iomp5.{shlib_ext}']]
        libraries = [os.path.join(libprefix, f) for f in libfiles]

        custom_paths = {
            'files': binaries + libraries,
            'dirs': [],
        }

        # make very sure that expected 'compilers_and_libraries_<VERSION>/linux' subdir is there for recent versions,
        # since we rely on it being there in module_load_environment
        if self.comp_libs_subdir:
            custom_paths['dirs'].append(self.comp_libs_subdir)

        custom_commands = ["which ifort"]

        IntelBase.sanity_check_step(self, custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2018-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing a bundle of Python packages, implemented as a generic easyblock

@author: Kenneth Hoste (Ghent University)
"""
import os

from easybuild.easyblocks.generic.bundle import Bundle
from easybuild.easyblocks.generic.pythonpackage import EXTS_FILTER_PYTHON_PACKAGES, run_pip_check, set_py_env_vars
from easybuild.easyblocks.generic.pythonpackage import PythonPackage, get_pylibdirs, find_python_cmd_from_ec
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import build_option, PYTHONPATH, EBPYTHONPREFIXES
from easybuild.tools.modules import get_software_root


class PythonBundle(Bundle):
    """
    Bundle of PythonPackages: install Python packages as extensions in a bundle
    Defines custom sanity checks and module environment
    """

    @staticmethod
    def extra_options(extra_vars=None):
        """Easyconfig parameters specific to bundles of Python packages."""
        if extra_vars is None:
            extra_vars = {}
        # combine custom easyconfig parameters of Bundle & PythonPackage
        extra_vars = Bundle.extra_options(extra_vars)
        extra_vars['default_easyblock'][0] = 'PythonPackage'
        return PythonPackage.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialize PythonBundle easyblock."""
        super().__init__(*args, **kwargs)

        self.cfg['exts_defaultclass'] = 'PythonPackage'
        self.cfg['exts_filter'] = EXTS_FILTER_PYTHON_PACKAGES

        # need to disable templating to ensure that actual value for exts_default_options is updated...
        with self.cfg.disable_templating():
            # set default options for extensions according to relevant top-level easyconfig parameters
            pypkg_keys = PythonPackage.extra_options().keys()
            for key in pypkg_keys:
                if key not in self.cfg['exts_default_options']:
                    self.cfg['exts_default_options'][key] = self.cfg[key]

            self.log.info("exts_default_options: %s", self.cfg['exts_default_options'])

        self.python_cmd = None
        self.pylibdir = None
        self.all_pylibdirs = None

        # figure out whether this bundle of Python packages is being installed for multiple Python versions
        self.multi_python = 'Python' in self.cfg['multi_deps']

    def prepare_python(self):
        """Python-specific preparations."""

        if get_software_root('Python') is None:
            raise EasyBuildError("Python not included as dependency!")
        self.python_cmd = find_python_cmd_from_ec(self.log, self.cfg, required=True)

        self.all_pylibdirs = get_pylibdirs(python_cmd=self.python_cmd)
        self.pylibdir = self.all_pylibdirs[0]

        # if 'python' is not used, we need to take that into account in the extensions filter
        # (which is also used during the sanity check)
        if self.python_cmd != 'python':
            orig_exts_filter = EXTS_FILTER_PYTHON_PACKAGES
            self.cfg['exts_filter'] = (orig_exts_filter[0].replace('python', self.python_cmd), orig_exts_filter[1])

    def prepare_step(self, *args, **kwargs):
        """Prepare for installing bundle of Python packages."""
        super().prepare_step(*args, **kwargs)
        self.prepare_python()

    def extensions_step(self, *args, **kwargs):
        """Install extensions (usually PythonPackages)"""
        set_py_env_vars(self.log)
        super().extensions_step(*args, **kwargs)

    def test_step(self):
        """No global test step for bundle of Python packages."""
        # required since runtest is set to True for Python packages by default
        pass

    def make_module_extra(self, *args, **kwargs):
        """Extra statements to include in module file: update $PYTHONPATH."""
        txt = super().make_module_extra(*args, **kwargs)

        # update $EBPYTHONPREFIXES rather than $PYTHONPATH
        # if this Python package was installed for multiple Python versions, or if we prefer it
        use_ebpythonprefixes = False
        runtime_deps = [dep['name'] for dep in self.cfg.dependencies(runtime_only=True)]

        if 'Python' in runtime_deps:
            self.log.info("Found Python runtime dependency, so considering $EBPYTHONPREFIXES...")
            if build_option('prefer_python_search_path') == EBPYTHONPREFIXES:
                self.log.info("Preferred Python search path is $EBPYTHONPREFIXES, so using that")
                use_ebpythonprefixes = True

        if self.multi_python or use_ebpythonprefixes:
            path = ''  # EBPYTHONPREFIXES are relative to the install dir
            if path not in self.module_generator.added_paths_per_key[EBPYTHONPREFIXES]:
                txt += self.module_generator.prepend_paths(EBPYTHONPREFIXES, path)
        else:

            # the temporary module file that is generated before installing extensions
            # must add all subdirectories to $PYTHONPATH without checking existence,
            # otherwise paths will be missing since nothing is there initially
            if self.current_step == 'extensions':
                new_pylibdirs = self.all_pylibdirs
            else:
                new_pylibdirs = [
                    lib_dir for lib_dir in self.all_pylibdirs
                    if os.path.exists(os.path.join(self.installdir, lib_dir))
                ]

            for pylibdir in new_pylibdirs:
                if pylibdir not in self.module_generator.added_paths_per_key[PYTHONPATH]:
                    txt += self.module_generator.prepend_paths(PYTHONPATH, pylibdir)

        return txt

    def load_module(self, *args, **kwargs):
        """(Re)set environment variables after loading module file.

        Required here to ensure the variables are also defined for stand-alone installations,
        because the environment is reset to the initial environment right before loading the module.
        """

        super().load_module(*args, **kwargs)
        # Required here to ensure that it is defined for sanity check commands of the bundle
        # because the environment is reset to the initial environment right before loading the module
        set_py_env_vars(self.log)

    def sanity_check_step(self, *args, **kwargs):
        """Custom sanity check for bundle of Python package."""

        if self.pylibdir is None:
            # Python attributes not set up yet, happens e.g. with --sanity-check-only, so do it now.
            # This also ensures the exts_filter option for extensions is set correctly.
            # Load module first to get the right python command.
            if not self.sanity_check_module_loaded:
                self.sanity_check_load_module(extension=kwargs.get('extension', False),
                                              extra_modules=kwargs.get('extra_modules', None))
            self.prepare_python()

        # inject directory path that uses %(pyshortver)s template into default value for sanity_check_paths
        # this is relevant for installations of Python bundles for multiple Python versions (via multi_deps)
        # (we can not pass this via custom_paths, since then the %(pyshortver)s template value will not be resolved)
        if not self.cfg['sanity_check_paths']:
            self.cfg['sanity_check_paths'] = {
                'files': [],
                'dirs': [os.path.join('lib', 'python%(pyshortver)s', 'site-packages')],
            }

        super().sanity_check_step(*args, **kwargs)

    def _sanity_check_step_extensions(self):
        """Run the pip check for extensions if enabled"""
        super(PythonBundle, self)._sanity_check_step_extensions()

        sanity_pip_check = self.cfg['sanity_pip_check']
        unversioned_packages = set(self.cfg['unversioned_packages'])

        # The options should be set in the main EC and cannot be different between extensions.
        # For backwards compatibility and to avoid surprises enable the pip-check if it is enabled
        # in the main EC or any extension and build the union of all unversioned_packages.
        has_sanity_pip_check_mismatch = False
        all_unversioned_packages = unversioned_packages.copy()
        for ext in self.ext_instances:
            if isinstance(ext, PythonPackage):
                if ext.cfg['sanity_pip_check'] != sanity_pip_check:
                    has_sanity_pip_check_mismatch = True
                all_unversioned_packages.update(ext.cfg['unversioned_packages'])

        if has_sanity_pip_check_mismatch:
            self.log.deprecated("For bundles of PythonPackage extensions the sanity_pip_check parameter "
                                "must be set at the top level, outside of exts_list", '6.0')
            sanity_pip_check = True  # Either the main set it or any extension enabled it
        if all_unversioned_packages != unversioned_packages:
            self.log.deprecated("For bundles of PythonPackage extensions the unversioned_packages parameter "
                                "must be set at the top level, outside of exts_list", '6.0')

        if sanity_pip_check:
            run_pip_check(python_cmd=self.python_cmd, unversioned_packages=all_unversioned_packages)



##
# Copyright 2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing Term::ReadLine::Gnu.

@author: Alexander Grund (TU Dresden)
"""

from easybuild.easyblocks.generic.perlmodule import PerlModule
from easybuild.tools.modules import get_software_root


class EB_Term_colon__colon_ReadLine_colon__colon_Gnu(PerlModule):
    """Support for installing the Term::ReadLine::Gnu Perl module."""

    def __init__(self, *args, **kwargs):
        """Set configopts for dependencies"""
        super().__init__(*args, **kwargs)
        # Use the custom --prefix option to pass the installation prefixes of all direct dependencies
        # to avoid it picking up system libraries.
        prefix = ':'.join(get_software_root(dep['name']) for dep in self.cfg.dependencies())
        self.cfg.update('configopts', f"--prefix='{prefix}'")



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for software that is configured with CMake, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Ward Poelmans (Ghent University)
@author: Maxime Boissonneault (Compute Canada - Universite Laval)
"""
import glob
import re
import os
from easybuild.tools import LooseVersion

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import BUILD, CUSTOM
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.config import build_option
from easybuild.tools.filetools import change_dir, create_unused_dir, mkdir, read_file, which
from easybuild.tools.environment import setvar
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.systemtools import get_shared_lib_ext
from easybuild.tools.utilities import nub
import easybuild.tools.toolchain as toolchain


DEFAULT_CONFIGURE_CMD = 'cmake'


def det_cmake_version():
    """
    Determine active CMake version.
    """
    cmake_version = get_software_version('CMake')
    if cmake_version is None:
        # also take into account release candidate versions
        regex = re.compile(r"^[cC][mM]ake version (?P<version>[0-9]\.[0-9a-zA-Z.-]+)$", re.M)

        cmd = "cmake --version"
        cmd_res = run_shell_cmd(cmd, hidden=True, fail_on_error=False)
        out = cmd_res.output
        res = regex.search(out)
        if res:
            cmake_version = res.group('version')
        else:
            raise EasyBuildError("Failed to determine CMake version from output of '%s': %s", cmd, out)

    return cmake_version


def setup_cmake_env(tc):
    """Setup env variables that cmake needs in an EasyBuild context."""

    # Set the search paths for CMake
    tc_ipaths = tc.get_variable("CPPFLAGS", list)
    tc_lpaths = tc.get_variable("LDFLAGS", list)
    cpaths = os.getenv('CPATH', '').split(os.pathsep)
    lpaths = os.getenv('LD_LIBRARY_PATH', '').split(os.pathsep)
    include_paths = os.pathsep.join(nub(tc_ipaths + cpaths))
    library_paths = os.pathsep.join(nub(tc_lpaths + lpaths))
    setvar("CMAKE_INCLUDE_PATH", include_paths)
    setvar("CMAKE_LIBRARY_PATH", library_paths)


def setup_cmake_env_python_hints(cmake_version=None):
    """Set environment variables as hints for CMake to prefer the Python module, if loaded.
    Useful when there is no way to specify arguments for CMake directly,
    e.g. when CMake is called from within another build system.
    Otherwise get_cmake_python_config_[str/dict] should be used instead.
    """
    if cmake_version is None:
        cmake_version = det_cmake_version()
    if LooseVersion(cmake_version) < '3.12':
        raise EasyBuildError("Setting Python hints for CMake requires CMake version 3.12 or newer")
    python_root = get_software_root('Python')
    if python_root:
        python_version = LooseVersion(get_software_version('Python'))
        setvar('Python_ROOT_DIR', python_root)
        if python_version >= "3":
            setvar('Python3_ROOT_DIR', python_root)
        else:
            setvar('Python2_ROOT_DIR', python_root)


def get_cmake_python_config_dict():
    """Get a dictionary with CMake configuration options for finding Python if loaded as a module."""
    options = {}
    python_root = get_software_root('Python')
    if python_root:
        python_version = LooseVersion(get_software_version('Python'))
        python_exe = os.path.join(python_root, 'bin', 'python')
        # This is required for (deprecated) `find_package(PythonInterp ...)`
        options['PYTHON_EXECUTABLE'] = python_exe
        # Ensure that both `find_package(Python) and find_package(Python2/3)` work as intended
        options['Python_EXECUTABLE'] = python_exe
        if python_version >= "3":
            options['Python3_EXECUTABLE'] = python_exe
        else:
            options['Python2_EXECUTABLE'] = python_exe
    return options


def get_cmake_python_config_str():
    """Get CMake configuration arguments for finding Python if loaded as a module.
    This string is intended to be passed to the invocation of `cmake`.
    """
    options = get_cmake_python_config_dict()
    return ' '.join('-D%s=%s' % (key, value) for key, value in options.items())


class CMakeMake(ConfigureMake):
    """Support for configuring build with CMake instead of traditional configure script"""

    @staticmethod
    def extra_options(extra_vars=None):
        """Define extra easyconfig parameters specific to CMakeMake."""
        extra_vars = ConfigureMake.extra_options(extra_vars)
        extra_vars.update({
            'abs_path_compilers': [False, "Specify compilers via absolute file path (not via command names)", CUSTOM],
            'allow_system_boost': [False, "Always allow CMake to pick up on Boost installed in OS "
                                          "(even if Boost is included as a dependency)", CUSTOM],
            'build_shared_libs': [None, "Build shared library (instead of static library)"
                                        "None can be used to add no flag (usually results in static library)", CUSTOM],
            'build_type': [None, "Build type for CMake, e.g. Release."
                                 "Defaults to 'Release', 'RelWithDebInfo' or 'Debug' depending on "
                                 "toolchainopts[debug,noopt]", CUSTOM],
            'configure_cmd': [DEFAULT_CONFIGURE_CMD, "Configure command to use", CUSTOM],
            'generator': [None, "Build file generator to use. None to use CMakes default", CUSTOM],
            'install_target_subdir': [None, "Subdirectory to use as installation target", CUSTOM],
            'install_libdir': ['lib', "Subdirectory to use for library installation files", CUSTOM],
            'runtest': [None, "Make target to test build or True to use CTest", BUILD],
            'srcdir': [None, "Source directory location to provide to cmake command", CUSTOM],
            'separate_build_dir': [True, "Perform build in a separate directory", CUSTOM],
        })
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Constructor for CMakeMake easyblock"""
        super().__init__(*args, **kwargs)
        self._lib_ext = None
        self._cmake_version = None
        self.separate_build_dir = None

    @property
    def lib_ext(self):
        """Return the extension for libraries build based on `build_shared_libs` or None if that is unset"""
        if self._lib_ext is None:
            build_shared_libs = self.cfg.get('build_shared_libs')
            if build_shared_libs:
                self._lib_ext = get_shared_lib_ext()
            elif build_shared_libs is not None:
                self._lib_ext = 'a'
        return self._lib_ext

    @lib_ext.setter
    def lib_ext(self, value):
        self._lib_ext = value

    @property
    def cmake_version(self):
        """Return the used CMake version, caching the value for reuse"""
        if self._cmake_version is None:
            self._cmake_version = det_cmake_version()
            self.log.debug('Determined CMake version: %s', self._cmake_version)
        return self._cmake_version

    @property
    def build_type(self):
        """Build type set in the EasyConfig with default determined by toolchainopts"""
        build_type = self.cfg.get('build_type')
        if build_type is None:
            if self.toolchain.options.get('noopt', None):  # also implies debug but is the closest match
                build_type = 'Debug'
            elif self.toolchain.options.get('debug', None):
                build_type = 'RelWithDebInfo'
            else:
                build_type = 'Release'
        return build_type

    def prepend_config_opts(self, config_opts):
        """Prepends configure options (-Dkey=value) to configopts ignoring those already set"""
        # need to disable template resolution or it will remain the same for all runs
        with self.cfg.disable_templating():
            cfg_configopts = self.cfg['configopts']

        # All options are of the form '-D<key>=<value>'
        new_opts = ' '.join('-D%s=%s' % (key, value) for key, value in config_opts.items()
                            if '-D%s=' % key not in cfg_configopts)
        self.cfg['configopts'] = ' '.join([new_opts, cfg_configopts])

    def configure_step(self, srcdir=None, builddir=None):
        """Configure build using cmake"""

        setup_cmake_env(self.toolchain)

        if builddir is None and self.cfg.get('separate_build_dir', True):
            self.separate_build_dir = create_unused_dir(self.builddir, 'easybuild_obj')
            builddir = self.separate_build_dir

        if builddir:
            mkdir(builddir, parents=True)
            change_dir(builddir)
            default_srcdir = self.cfg['start_dir']
        else:
            default_srcdir = '.'

        if srcdir is None:
            if self.cfg.get('srcdir', None) is not None:
                # Note that the join returns srcdir if it is absolute
                srcdir = os.path.join(default_srcdir, self.cfg['srcdir'])
            else:
                srcdir = default_srcdir

        install_target = self.installdir
        install_target_subdir = self.cfg.get('install_target_subdir')
        if install_target_subdir:
            install_target = os.path.join(install_target, install_target_subdir)
        options = {'CMAKE_INSTALL_PREFIX': install_target}

        if self.installdir.startswith('/opt') or self.installdir.startswith('/usr'):
            # https://cmake.org/cmake/help/latest/module/GNUInstallDirs.html
            localstatedir = os.path.join(self.installdir, 'var')
            runstatedir = os.path.join(localstatedir, 'run')
            sysconfdir = os.path.join(self.installdir, 'etc')
            options['CMAKE_INSTALL_LOCALSTATEDIR'] = localstatedir
            options['CMAKE_INSTALL_RUNSTATEDIR'] = runstatedir
            options['CMAKE_INSTALL_SYSCONFDIR'] = sysconfdir

        if '-DCMAKE_BUILD_TYPE=' in self.cfg['configopts']:
            if self.cfg.get('build_type') is not None:
                self.log.info("CMAKE_BUILD_TYPE is set in configopts. Ignoring 'build_type' easyconfig parameter.")
        else:
            options['CMAKE_BUILD_TYPE'] = self.build_type

        # Set installation directory for libraries
        # any CMAKE_INSTALL_DIR[:PATH] setting defined in 'configopts' has precedence over 'install_libdir'
        if self.cfg['install_libdir'] is not None:
            cmake_install_dir_pattern = re.compile(r"-DCMAKE_INSTALL_LIBDIR(:PATH)?=[^\s]")
            if cmake_install_dir_pattern.search(self.cfg['configopts']):
                self.log.info(
                    "CMAKE_INSTALL_LIBDIR is set in configopts. Ignoring 'install_libdir' easyconfig parameter."
                )
            else:
                # set CMAKE_INSTALL_LIBDIR including its type to PATH, otherwise CMake can silently ignore it
                options['CMAKE_INSTALL_LIBDIR:PATH'] = self.cfg['install_libdir']

        # Add -fPIC flag if necessary
        if self.toolchain.options.get('pic', False):
            options['CMAKE_POSITION_INDEPENDENT_CODE'] = 'ON'

        if self.cfg['generator']:
            generator = '-G "%s"' % self.cfg['generator']
        else:
            generator = ''

        # pass --sysroot value down to CMake,
        # and enable using absolute paths to compiler commands to avoid
        # that CMake picks up compiler from sysroot rather than toolchain compiler...
        sysroot = build_option('sysroot')
        if sysroot:
            options['CMAKE_SYSROOT'] = sysroot
            self.log.info("Using absolute path to compiler commands because of alterate sysroot %s", sysroot)
            self.cfg['abs_path_compilers'] = True

        # Set flag for shared libs if requested
        # Not adding one allows the project to choose a default
        build_shared_libs = self.cfg.get('build_shared_libs')
        if build_shared_libs is not None:
            # Contrary to other options build_shared_libs takes precedence over configopts which may be unexpected.
            # This is to allow self.lib_ext to be determined correctly.
            # Usually you want to remove -DBUILD_SHARED_LIBS from configopts and set build_shared_libs to True or False
            # If you need it in configopts don't set build_shared_libs (or explicitely set it to `None` (Default))
            if '-DBUILD_SHARED_LIBS=' in self.cfg['configopts']:
                print_warning('Ignoring BUILD_SHARED_LIBS setting in configopts because build_shared_libs is set')
            self.cfg.update('configopts', '-DBUILD_SHARED_LIBS=%s' % ('ON' if build_shared_libs else 'OFF'))

        # If the cache does not exist CMake reads the environment variables
        cache_exists = os.path.exists('CMakeCache.txt')
        env_to_options = dict()

        # Setting compilers is not required unless we want absolute paths
        if self.cfg.get('abs_path_compilers', False) or cache_exists:
            env_to_options.update({
                'CC': 'CMAKE_C_COMPILER',
                'CXX': 'CMAKE_CXX_COMPILER',
                'F90': 'CMAKE_Fortran_COMPILER',
            })
        else:
            # Set the variable which CMake uses to init the compiler using F90 for backward compatibility
            fc = os.getenv('F90')
            if fc:
                setvar('FC', fc)

        # Flags are read from environment variables already since at least CMake 2.8.0
        if LooseVersion(self.cmake_version) < LooseVersion('2.8.0') or cache_exists:
            env_to_options.update({
                'CFLAGS': 'CMAKE_C_FLAGS',
                'CXXFLAGS': 'CMAKE_CXX_FLAGS',
                'FFLAGS': 'CMAKE_Fortran_FLAGS',
            })

        for env_name, option in env_to_options.items():
            value = os.getenv(env_name)
            if value is not None:
                if option.endswith('_COMPILER') and self.cfg.get('abs_path_compilers', False):
                    value = which(value)
                    self.log.info("Using absolute path to compiler command: %s", value)
                options[option] = value

        if build_option('rpath') and LooseVersion(self.cmake_version) < LooseVersion('3.5.0'):
            # instruct CMake not to fiddle with RPATH when --rpath is used, since it will undo stuff on install...
            # this is only required for CMake < 3.5.0, since newer version are more careful w.r.t. RPATH,
            # see https://github.com/Kitware/CMake/commit/3ec9226779776811240bde88a3f173c29aa935b5
            options['CMAKE_SKIP_RPATH'] = 'ON'

        # show what CMake is doing by default
        options['CMAKE_VERBOSE_MAKEFILE'] = 'ON'

        # disable CMake user package repository
        options['CMAKE_FIND_USE_PACKAGE_REGISTRY'] = 'OFF'

        # ensure CMake uses EB python, not system or virtualenv python
        options.update(get_cmake_python_config_dict())

        # pass the preferred host compiler, CUDA compiler, and CUDA architectures to the CUDA compiler
        cuda_root = get_software_root('CUDA')
        if cuda_root:
            options['CMAKE_CUDA_HOST_COMPILER'] = which(os.getenv('CXX', 'g++'))
            options['CMAKE_CUDA_COMPILER'] = which('nvcc')
            cuda_cc = build_option('cuda_compute_capabilities') or self.cfg['cuda_compute_capabilities']
            if cuda_cc:
                options['CMAKE_CUDA_ARCHITECTURES'] = '"%s"' % ';'.join([cc.replace('.', '') for cc in cuda_cc])
            else:
                raise EasyBuildError('List of CUDA compute capabilities must be specified, either via '
                                     'cuda_compute_capabilities easyconfig parameter or via '
                                     '--cuda-compute-capabilities')

        if not self.cfg.get('allow_system_boost', False):
            boost_root = get_software_root('Boost')
            if boost_root:
                # Check for older builds of Boost
                cmake_files = glob.glob(os.path.join(boost_root, 'lib', 'cmake', 'boost_system*',
                                                     'libboost_system-variant*-shared.cmake'))
                cmake_files = [os.path.basename(x) for x in cmake_files]
                if len(cmake_files) > 1 and 'libboost_system-variant-shared.cmake' in cmake_files:
                    # disable search for Boost CMake package configuration files when conflicting variant configs
                    # are present (builds using the old EasyBlock)
                    options['Boost_NO_BOOST_CMAKE'] = 'ON'

                # Don't pick up on system Boost if Boost is included as dependency
                # - specify Boost location via -DBOOST_ROOT
                # - instruct CMake to not search for Boost headers/libraries in other places
                options['BOOST_ROOT'] = boost_root
                options['Boost_NO_SYSTEM_PATHS'] = 'ON'

        self.cmake_options = options

        if self.cfg.get('configure_cmd') == DEFAULT_CONFIGURE_CMD:
            self.prepend_config_opts(options)
            command = ' '.join([
                self.cfg['preconfigopts'],
                DEFAULT_CONFIGURE_CMD,
                generator,
                self.cfg['configopts'],
                srcdir])
        else:
            command = ' '.join([
                self.cfg['preconfigopts'],
                self.cfg.get('configure_cmd'),
                self.cfg['configopts']])

        res = run_shell_cmd(command)
        self.check_python_paths()
        return res.output

    def check_python_paths(self):
        """Check that there are no detected Python paths outside the Python dependency provided by EasyBuild"""
        if not os.path.exists('CMakeCache.txt'):
            self.log.warning("CMakeCache.txt not found. Python paths checks skipped.")
            return
        cmake_cache = read_file('CMakeCache.txt')
        if not cmake_cache:
            self.log.warning("CMake Cache could not be read. Python paths checks skipped.")
            return

        self.log.info("Checking Python paths")

        python_paths = {
            "executable": [],
            "include_dir": [],
            "library": [],
        }
        python_regex = re.compile(r"_?(Python|PYTHON)\d?_"
                                  r"(?P<type>EXECUTABLE|INCLUDE_DIR|LIBRARY)\w*(:\w+)?\s*=(?P<value>.*)")
        cmake_false_expressions = {'', '0', 'OFF', 'NO', 'FALSE', 'N', 'IGNORE', 'NOTFOUND'}
        for line in cmake_cache.splitlines():
            match = python_regex.match(line)
            if match:
                self.log.debug("Python related CMake cache line found: " + line)
                path_type = match['type'].lower()
                path = match['value'].strip()
                if path.endswith('-NOTFOUND') or path.upper() in cmake_false_expressions:
                    continue
                self.log.info("Python %s path: %s", path_type, path)
                python_paths[path_type].append(path)

        ebrootpython_path = get_software_root("Python")
        if not ebrootpython_path:
            if any(python_paths.values()) and not self.toolchain.comp_family() == toolchain.SYSTEM:
                self.log.warning("Found Python paths in CMake cache but Python is not a dependency")
            # Can't do the check
            return
        ebrootpython_path = os.path.realpath(ebrootpython_path)

        errors = []
        for path_type, paths in python_paths.items():
            for path in paths:
                if not os.path.exists(path):
                    errors.append("Python %s path does not exist: %s" % (path_type, path))
                elif not os.path.realpath(path).startswith(ebrootpython_path):
                    errors.append("Python %s path '%s' is outside EBROOTPYTHON (%s)" %
                                  (path_type, path, ebrootpython_path))

        if errors:
            # Combine all errors into a single message
            error_message = "\n".join(errors)
            raise EasyBuildError("Python path errors:\n" + error_message)
        self.log.info("Python check successful")

    def test_step(self):
        """CMake specific test setup"""
        # When using ctest for tests (default) then show verbose output if a test fails
        setvar('CTEST_OUTPUT_ON_FAILURE', 'True')
        # Handle `runtest = True` if `test_cmd` has not been set
        if self.cfg.get('runtest') is True and not self.cfg.get('test_cmd'):
            test_cmd = 'ctest'
            if LooseVersion(self.cmake_version) >= '3.17.0':
                test_cmd += ' --no-tests=error'
            self.log.debug("`runtest = True` found, using '%s' as test_cmd", test_cmd)
            self.cfg['test_cmd'] = test_cmd

        super().test_step()



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing FSL, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""

import difflib
import os
import re
from easybuild.tools import LooseVersion

import easybuild.tools.environment as env
from easybuild.framework.easyblock import EasyBlock
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.filetools import read_file, copy_dir
from easybuild.tools.utilities import nub


class EB_FSL(EasyBlock):
    """Support for building and installing FSL."""

    def __init__(self, *args, **kwargs):
        """Specify building in install dir, initialize custom variables."""

        super().__init__(*args, **kwargs)

        self.build_in_installdir = True

        self.fsldir = None

        self.module_load_environment.PATH = [os.path.join('fsl', 'bin')]
        self.module_load_environment.LD_LIBRARY_PATH = [os.path.join('fsl', 'lib')]

    def configure_step(self):
        """Configure FSL build: set FSLDIR env var."""

        self.fsldir = self.cfg['start_dir']
        env.setvar('FSLDIR', self.fsldir)

        # determine FSL machine type
        cmd = ". %s/etc/fslconf/fslmachtype.sh" % self.fsldir
        res = run_shell_cmd(cmd)
        fslmachtype = res.output.strip()
        self.log.debug("FSL machine type: %s" % fslmachtype)

        best_cfg = None

        # Predefined makefiles for various configs have disappeared in v6.0.2.
        # The next part of the EasyBlock has thereby become obsolete.
        # See https://github.com/easybuilders/easybuild-easyblocks/issues/1859
        if LooseVersion(self.version) < LooseVersion('6.0.2'):

            # Patch files for ver. < 5.0.10 patch multiple config directories
            if LooseVersion(self.version) >= LooseVersion('5.0.10'):
                # Check if a specific machine type directory is patched
                systype_regex = re.compile(r"^diff.*config\/(.*(apple|gnu|i686|linux|spark)(?:(?!\/).)*)", re.M)

                patched_cfgs = []

                for patch in self.patches:
                    patchfile = read_file(patch['path'])
                    res = systype_regex.findall(patchfile)
                    patched_cfgs.extend([i[0] for i in res])

                # Check that at least one config has been found
                if patched_cfgs:
                    # Check that a single config has been patched
                    if len(nub(patched_cfgs)) == 1:
                        best_cfg = patched_cfgs[0]
                        self.log.debug("Found patched config dir: %s", best_cfg)
                    else:
                        raise EasyBuildError("Patch files are editing multiple config dirs: %s", patched_cfgs)
                else:
                    self.log.debug("No config dir found in patch files")

            # If no patched config is found, pick best guess
            cfgdir = os.path.join(self.fsldir, "config")
            try:
                if not best_cfg:
                    cfgs = os.listdir(cfgdir)
                    best_cfg = difflib.get_close_matches(fslmachtype, cfgs)[0]
                    self.log.debug("Best matching config dir for %s is %s" % (fslmachtype, best_cfg))
            except OSError as err:
                raise EasyBuildError("Unable to access configuration directory: %s", cfgdir, err)

            # Prepare config
            # Either use patched config or copy closest match
            if fslmachtype != best_cfg:
                srcdir = os.path.join(cfgdir, best_cfg)
                tgtdir = os.path.join(cfgdir, fslmachtype)
                copy_dir(srcdir, tgtdir)
                self.log.debug("Copied %s to %s" % (srcdir, tgtdir))

    def build_step(self):
        """Build FSL using supplied script."""

        cmd = ". %s/etc/fslconf/fsl.sh && ./build" % self.fsldir
        run_shell_cmd(cmd)

        # check build.log file for success
        buildlog = os.path.join(self.installdir, "fsl", "build.log")
        f = open(buildlog, "r")
        txt = f.read()
        f.close()

        error_regexp = re.compile("ERROR in BUILD")
        if error_regexp.search(txt):
            raise EasyBuildError("Error detected in build log %s.", buildlog)

    def install_step(self):
        """Building was performed in install dir, no explicit install step required."""
        pass

    def make_module_extra(self):
        """Add setting of FSLDIR in module."""

        txt = super().make_module_extra()

        txt += self.module_generator.set_environment("FSLDIR", os.path.join(self.installdir, 'fsl'))

        return txt

    def sanity_check_step(self):
        """Custom sanity check for FSL"""

        custom_paths = {
            'files': [],
            'dirs': ['fsl/%s' % x for x in ['bin', 'data', 'etc', 'extras', 'include', 'lib']],
        }

        super().sanity_check_step(custom_paths=custom_paths)



# #
# Copyright 2013-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
EasyBuild support for installing Intel Inspector, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Damian Alvarez (Forschungzentrum Juelich GmbH)
"""
import os
from easybuild.tools import LooseVersion

from easybuild.easyblocks.generic.intelbase import IntelBase
from easybuild.tools.build_log import EasyBuildError


class EB_Inspector(IntelBase):
    """
    Support for installing Intel Inspector
    - minimum version suported: 2020.x
    """

    def __init__(self, *args, **kwargs):
        """Easyblock constructor; define class variables."""
        super().__init__(*args, **kwargs)

        loosever = LooseVersion(self.version)
        if loosever < LooseVersion('2020'):
            raise EasyBuildError(
                f"Version {self.version} of {self.name} is unsupported. Mininum supported version is 2020.0."
            )

        # recent versions of Inspector are installed to a subdirectory
        self.subdir = ''
        if loosever < LooseVersion('2021'):
            self.subdir = 'inspector'
        elif loosever >= LooseVersion('2021'):
            self.subdir = os.path.join('inspector', 'latest')

        # prepare module load environment
        self.prepare_intel_tools_env()

    def make_installdir(self):
        """Do not create installation directory, install script handles that already."""
        super().make_installdir(dontcreate=True)

    def sanity_check_step(self):
        """Custom sanity check paths for Intel Inspector."""
        binaries = ['inspxe-cl', 'inspxe-feedback', 'inspxe-gui', 'inspxe-runmc', 'inspxe-runtc']
        custom_paths = self.get_custom_paths_tools(binaries)
        super().sanity_check_step(custom_paths=custom_paths)



#! /usr/bin/env python
##
# Copyright 2020-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing PyTorch, implemented as an easyblock

@author: Alexander Grund (TU Dresden)
"""

import os
import re
import sys
import tempfile
import xml.etree.ElementTree as ET
from collections import Counter
from enum import Enum
from itertools import chain, groupby
from operator import attrgetter
from pathlib import Path
from typing import Dict, Iterable, List

import easybuild.tools.environment as env
from easybuild.easyblocks.generic.pythonpackage import PythonPackage
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools import LooseVersion
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.config import ERROR, build_option
from easybuild.tools.filetools import apply_regex_substitutions, mkdir, symlink
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.systemtools import POWER, get_cpu_architecture

if sys.version_info >= (3, 9):
    from dataclasses import dataclass

    @dataclass
    class FailedTestNames:
        """Hold list of tests names that failed with error or failure"""
        error: list[str]
        fail: list[str]

    @dataclass
    class TestSuiteResult:
        """Hold the name of a test suite and a summary of the failures"""
        name: str
        summary: str

    @dataclass
    class TestResult:
        """Status report and results of a test run"""
        test_cnt: int
        error_cnt: int
        failure_cnt: int
        failed_suites: list[TestSuiteResult]
        terminated_suites: dict[str, str]  # Name and signal of terminated suites
        all_failed_suites: set[str]  # Names of all failed suites
else:
    from collections import namedtuple
    FailedTestNames = namedtuple('FailedTestNames', ('error', 'fail'))
    TestSuiteResult = namedtuple('TestSuiteResult', ('name', 'summary'))
    TerminatedTestSuite = namedtuple('TerminatedTestSuite', ('name', 'signal'))
    TestResult = namedtuple('TestResult', ('test_cnt',
                                           'error_cnt',
                                           'failure_cnt',
                                           'failed_suites',
                                           'terminated_suites',
                                           'all_failed_suites'
                                           ))


def find_failed_test_names(tests_out):
    """Find failed names of failed test cases in the output of the test step

    Return sorted list of names in FailedTestNames tuple
    """
    # patterns like
    # === FAIL: test_add_scalar_relu (quantization.core.test_quantized_op.TestQuantizedOps) ===
    # --- ERROR: test_all_to_all_group_cuda (__main__.TestDistBackendWithSpawn) ---
    regex = r"^[=-]+\n(FAIL|ERROR): (test_.*?)\s\(.*\n[=-]+\n"
    failed_test_cases = re.findall(regex, tests_out, re.M)
    # And patterns like:
    # FAILED test_ops_gradients.py::TestGradientsCPU::test_fn_grad_linalg_det_singular_cpu_complex128 - [snip]
    # FAILED [22.8699s] test_sparse_csr.py::TestSparseCompressedCPU::test_invalid_input_csr_large_cpu - [snip]
    # FAILED [0.0623s] dynamo/test_dynamic_shapes.py::DynamicShapesExportTests::test_predispatch -  [snip]
    regex = r"^(FAILED) (?:\[.*?\] )?(?:\w|/)+\.py.*::(test_.*?) - "
    failed_test_cases.extend(re.findall(regex, tests_out, re.M))
    return FailedTestNames(error=sorted(set(m[1] for m in failed_test_cases if m[0] == 'ERROR')),
                           fail=sorted(set(m[1] for m in failed_test_cases if m[0] != 'ERROR')))


def parse_test_log(tests_out):
    """Parse the test output and return result as TestResult tuple"""

    def get_count_for_pattern(regex, text):
        """Match the regexp containing a single group and return the integer value of the matched group.
            Return zero if no or more than 1 match was found and warn for the latter case
        """
        match = re.findall(regex, text)
        if len(match) == 1:
            return int(match[0])
        elif len(match) > 1:
            # Shouldn't happen, but means something went wrong with the regular expressions.
            # Throw warning, as the build might be fine, no need to error on this.
            warn_msg = "Error in counting the number of test failures in the output of the PyTorch test suite.\n"
            warn_msg += "Please check the EasyBuild log to verify the number of failures (if any) was acceptable."
            print_warning(warn_msg)
        return 0

    failure_cnt = 0
    error_cnt = 0
    failed_suites = []

    # Remove empty lines to make RegExs below simpler
    tests_out = re.sub(r'^[ \t]*\n', '', tests_out, flags=re.MULTILINE)

    # Examples: "test_jit_profiling failed! Received signal: SIGSEGV"
    #           "test_weak failed!"
    #           "test_decomp 1/1 failed!"
    #           "test_pytree 1/1 failed! [Errno 2] No such file or directory: '/dev/shm/build/...'"
    suite_failed_pattern = (r"^(?P<failed_test_suite_name>.*?) (?:\d+/\d+ )?failed!"
                            r"(?: Received signal: (\w+)| \[Errno \d+\] .*)?\s*$")

    # Grep for patterns like:
    # Ran 219 tests in 67.325s
    #
    # FAILED (errors=10, skipped=190, expected failures=6)
    # test_fx failed!
    regex = (r"^Ran (?P<test_cnt>[0-9]+) tests.*$\n"
             r"FAILED \((?P<failure_summary>.*)\)$\n"
             r"(?:^(?:(?!failed!).)*$\n){0,5}"
             + suite_failed_pattern)

    for m in re.finditer(regex, tests_out, re.M):
        # E.g. 'failures=3, errors=10, skipped=190, expected failures=6'
        failure_summary = m.group('failure_summary')
        total, test_suite = m.group('test_cnt', 'failed_test_suite_name')
        failed_suites.append(
            TestSuiteResult(test_suite, "{total} total tests, {failure_summary}".format(
                total=total, failure_summary=failure_summary))
        )
        failure_cnt += get_count_for_pattern(r"(?<!expected )failures=([0-9]+)", failure_summary)
        error_cnt += get_count_for_pattern(r"errors=([0-9]+)", failure_summary)

    # Grep for patterns like:
    # ===================== 2 failed, 128 passed, 2 skipped, 2 warnings in 3.43s =====================
    # test_quantization failed!
    # OR:
    # ===================== 2 failed, 128 passed, 2 skipped, 2 warnings in 63.43s (01:03:43) =========
    #
    # FINISHED PRINTING LOG FILE
    # test_quantization failed!
    # OR:
    # ===================== 2 failed, 128 passed, 2 skipped, 2 warnings in 63.43s (01:03:43) =========
    # If in CI, skip info is located in the xml test reports, please either go to s3 or the hud to download them
    #
    # FINISHED PRINTING LOG FILE of test_ops_gradients (/tmp/vsc40023/easybuil...)
    #
    # test_quantization failed!

    regex = (
        r"^=+ (?P<failure_summary>.*) in [0-9]+\.*[0-9]*[a-zA-Z]* (\([0-9]+:[0-9]+:[0-9]+\) )?=+$\n"
        r"(?:.*skip info is located in the xml test reports.*\n)?"
        r"(?:.*FINISHED PRINTING LOG FILE.*\n)?"
        + suite_failed_pattern
    )

    for m in re.finditer(regex, tests_out, re.M):
        # E.g. '2 failed, 128 passed, 2 skipped, 2 warnings'
        failure_summary = m.group('failure_summary')
        test_suite = m.group('failed_test_suite_name')
        failed_suites.append(TestSuiteResult(test_suite, failure_summary))
        failure_cnt += get_count_for_pattern(r"([0-9]+) failed", failure_summary)
        error_cnt += get_count_for_pattern(r"([0-9]+) error", failure_summary)

    # Grep for patterns like:
    # AssertionError: 2 unit test(s) failed:
    #         DistributedDataParallelTest.test_find_unused_parameters_kwarg_debug_detail
    #         DistributedDataParallelTest.test_find_unused_parameters_kwarg_grad_is_view_debug_detail
    #
    # FINISHED PRINTING LOG FILE of distributed/test_c10d_nccl (<snip>)
    #
    # distributed/test_c10d_nccl failed!

    regex = (
        r"^AssertionError: (?P<failure_summary>[0-9]+ unit test\(s\) failed):\n"
        r"(\s+.*\n)+"
        r"(((?!failed!).)*\n){0,5}"
        + suite_failed_pattern
    )

    for m in re.finditer(regex, tests_out, re.M):
        # E.g. '2 unit test(s) failed'
        failure_summary = m.group('failure_summary')
        test_suite = m.group('failed_test_suite_name')
        failed_suites.append(TestSuiteResult(test_suite, failure_summary))
        failure_cnt += get_count_for_pattern(r"([0-9]+) unit test\(s\) failed", failure_summary)

    # Collect total number of tests

    # Pattern for tests ran with unittest like:
    # Ran 3 tests in 0.387s
    regex = r"^Ran (?P<test_cnt>[0-9]+) tests in"
    test_cnt = sum(int(hit) for hit in re.findall(regex, tests_out, re.M))
    # Pattern for tests ran with pytest like:
    # ============ 286 passed, 18 skipped, 2 xfailed in 38.71s ============
    regex = r"=+ (?P<summary>.*) in \d+.* =+\n"
    count_patterns = [re.compile(r"([0-9]+) " + reason) for reason in [
        "failed",
        "passed",
        "skipped",
        "deselected",
        "xfailed",
        "xpassed",
    ]]
    for m in re.finditer(regex, tests_out, re.M):
        test_cnt += sum(get_count_for_pattern(p, m.group("summary")) for p in count_patterns)

    # Gather all failed tests suites in case we missed any,
    # e.g. when it exited due to syntax errors or with a signal such as SIGSEGV
    failed_suites_and_signal = set(re.findall(suite_failed_pattern, tests_out, re.M))

    return TestResult(test_cnt=test_cnt, error_cnt=error_cnt, failure_cnt=failure_cnt,
                      failed_suites=failed_suites,
                      # Assumes that the suite name is unique
                      terminated_suites={name: signal for name, signal in failed_suites_and_signal if signal},
                      all_failed_suites={i[0] for i in failed_suites_and_signal})


class EB_PyTorch(PythonPackage):
    """Support for building/installing PyTorch."""

    GENERATE_TEST_REPORT_VAR_NAME = 'EASYBUILD_WRITE_PYTORCH_TEST_REPORTS'

    @staticmethod
    def extra_options():
        extra_vars = PythonPackage.extra_options()
        extra_vars.update({
            'build_type': [None, "Build type for CMake, e.g. Release."
                                 "Defaults to 'Release' or 'Debug' depending on toolchainopts[debug]", CUSTOM],
            'custom_opts': [[], "List of options for the build/install command. Can be used to change the defaults " +
                                "set by the PyTorch EasyBlock, for example ['USE_MKLDNN=0'].", CUSTOM],
            'excluded_tests': [{}, "Mapping of architecture strings to list of tests to be excluded", CUSTOM],
            'max_failed_tests': [0, "Maximum number of failing tests", CUSTOM],
        })

        # disable use of pip to install PyTorch by default, overwriting the default set in PythonPackage;
        # see also https://github.com/easybuilders/easybuild-easyblocks/pull/3022
        extra_vars['use_pip'][0] = None

        # Make pip show output of build process as that may often contain errors or important warnings
        extra_vars['pip_verbose'][0] = True
        # Test as-if pytorch was installed
        extra_vars['testinstall'][0] = True

        return extra_vars

    def __init__(self, *args, **kwargs):
        """Constructor for PyTorch easyblock."""
        super().__init__(*args, **kwargs)
        self.options['modulename'] = 'torch'
        self.has_xml_test_reports = False

        self.tmpdir = tempfile.mkdtemp(suffix='-pytorch-build')

        # opt-in to using pip to install PyTorch for sufficiently recent version (>= 2.0),
        # unless it's otherwise specified
        pytorch_version = LooseVersion(self.version)
        if self.cfg['use_pip'] is None and pytorch_version >= '2.0':
            self.log.info("Auto-enabling use of pip to install PyTorch >= 2.0, since 'use_pip' is not set")
            self.cfg['use_pip'] = True
            self.determine_install_command()

        # Set extra environment variables for PyTorch
        # use glob pattern as self.pylibdir is unknown at this stage
        # it will be expanded before injection into the module file
        py_site_glob = os.path.join('lib', 'python*', 'site-packages')
        self.module_load_environment.CMAKE_PREFIX_PATH = [os.path.join(py_site_glob, 'torch')]
        # required to dynamically load libcaffe2_nvrtc.so
        self.module_load_environment.LD_LIBRARY_PATH = [os.path.join(py_site_glob, 'torch', 'lib')]
        # important when RPATH linking is enabled
        self.module_load_environment.LIBRARY_PATH = [os.path.join(py_site_glob, 'torch', 'lib')]

    def fetch_step(self, skip_checksums=False):
        """Fetch sources for installing PyTorch, including those for tests."""
        super().fetch_step(skip_checksums)
        # Resolve tests early to avoid failures later. Use obtain_file if path is not absolute
        tests = [test if os.path.isabs(test) else self.obtain_file(test) for test in self.cfg['tests']]
        self.cfg['tests'] = tests

    @staticmethod
    def get_dependency_options_for_version(pytorch_version):
        """
        PyTorch can enable some functionality based on available software or use system software instead of a submodule
        This returns EasyBuild names of that and the flag that should be used when the dependency is found

        The result is a list of tuples (enable_flag, eb_name)
        """
        pytorch_version = LooseVersion(pytorch_version)

        def is_version_ok(version_range):
            """Return True if the PyTorch version to be installed matches the version_range"""
            min_version, max_version = version_range.split(':')
            result = True
            if min_version and pytorch_version < min_version:
                result = False
            if max_version and pytorch_version >= max_version:
                result = False
            return result

        available_libs = (
            # Format: (PyTorch flag to enable, EB name, '<min version>:<exclusive max version>')
            # Use `None` for the EB name if no known EC exists
            ('USE_FFMPEG=1', 'FFmpeg', '1.0.0:'),
            ('USE_GFLAGS=1', 'gflags', '1.0.0:'),
            ('USE_GLOG=1', 'glog', '1.0.0:'),

            # For system libs check CMakeLists.txt, below `if(USE_SYSTEM_LIBS)`, order kept here
            # NCCL handled specially as other env variables are requires for it
            ('USE_SYSTEM_CPUINFO=1', None, '1.6.0:'),
            ('USE_SYSTEM_SLEEF=1', None, '1.6.0:'),
            ('USE_SYSTEM_GLOO=1', None, '1.6.0:'),
            ('BUILD_CUSTOM_PROTOBUF=0', 'protobuf', '1.2.0:'),
            ('USE_SYSTEM_EIGEN_INSTALL=1', 'Eigen', '1.0.0:'),
            ('USE_SYSTEM_FP16=1', None, '1.6.0:'),
            ('USE_SYSTEM_PTHREADPOOL=1', None, '1.6.0:'),
            ('USE_SYSTEM_PSIMD=1', None, '1.6.0:'),
            ('USE_SYSTEM_FXDIV=1', None, '1.6.0:'),
            ('USE_SYSTEM_BENCHMARK=1', None, '1.6.0:'),  # Google Benchmark
            ('USE_SYSTEM_ONNX=1', None, '1.6.0:'),
            ('USE_SYSTEM_PYBIND11=1', 'pybind11', '1.10.0:'),
            ('USE_SYSTEM_XNNPACK=1', None, '1.6.0:'),
        )
        return [(enable_opt, dep_name) for enable_opt, dep_name, version_range in available_libs
                if is_version_ok(version_range)]

    def prepare_step(self, *args, **kwargs):
        """Make sure that versioned CMake alias exists"""
        super().prepare_step(*args, **kwargs)
        # PyTorch preferes cmake3 over cmake which usually does not exist
        cmake_root = get_software_root('CMake')
        cmake_version = get_software_version('CMake')
        if cmake_root and not os.path.isfile(os.path.join(cmake_root, 'bin', 'cmake3')):
            if cmake_version and cmake_version.split('.')[0] != '3':
                raise EasyBuildError('PyTorch requires CMake 3 but CMake %s was found', cmake_version)
            cmake_bin_dir = tempfile.mkdtemp(suffix='cmake-bin')
            self.log.warning('Creating symlink `cmake3` in %s to avoid PyTorch picking up a system CMake. ' +
                             'Reinstall the CMake module to avoid this!', cmake_bin_dir)
            symlink(os.path.join(cmake_root, 'bin', 'cmake'), os.path.join(cmake_bin_dir, 'cmake3'))
            path = "%s:%s" % (cmake_bin_dir, os.getenv('PATH'))
            env.setvar('PATH', path)

    def configure_step(self):
        """Custom configure procedure for PyTorch."""
        super().configure_step()

        pytorch_version = LooseVersion(self.version)

        self.has_xml_test_reports = False
        if pytorch_version >= '1.10.0':
            res = run_shell_cmd(self.python_cmd + " -c 'import xmlrunner'", fail_on_error=False)
            if res.exit_code != 0:
                msg = ("Python package xmlrunner (unittest-xml-reporting) not found in dependencies of the "
                       "PyTorch EasyConfig, can't enable advanced test result checks. Output: " + res.output)
                # We introduced this in the 2.3 EasyConfig
                if pytorch_version >= '2.3':
                    print_warning(msg)
                else:
                    self.log.warning(msg)
            else:
                # Replace the condition to enable the XML test reports to use our variable instead of $IS_CI/$IS_IN_CI
                # The variable is used because the file gets installed and we shouldn't change the default behavior.
                apply_regex_substitutions('torch/testing/_internal/common_utils.py',
                                          [(r'(default=_get_test_report_path\(\) if) IS(_IN)?_CI else None',
                                            fr'\1 os.getenv("{self.GENERATE_TEST_REPORT_VAR_NAME}") else None')],
                                          backup=False, on_missing_match=ERROR)
                if pytorch_version >= '2.1.0':
                    run_test_subs = [(r'if IS_CI:\n\s+# Add the option to generate XML test report.*',
                                      'if TEST_SAVE_XML:\n')]
                else:
                    run_test_subs = [
                         (r'from torch.testing._internal.common_utils import\s+\(\n\s+',
                          r'\g<0>get_report_path, '),
                         (r'# If using pytest.*\n\s+if options.pytest:\n\s+unittest_args = \[',
                          r'\g<0>"--junit-xml-reruns", get_report_path(pytest=True)] + ['),
                    ]
                apply_regex_substitutions('test/run_test.py', run_test_subs, backup=False, on_missing_match=ERROR,
                                          single_line=False)
                self.has_xml_test_reports = True

        # Gather default options. Will be checked against (and can be overwritten by) custom_opts
        options = ['PYTORCH_BUILD_VERSION=' + self.version, 'PYTORCH_BUILD_NUMBER=1']

        def add_enable_option(name, enabled):
            """Add `name=0` or `name=1` depending on enabled"""
            options.append('%s=%s' % (name, '1' if enabled else '0'))

        # enable verbose mode when --debug is used (to show compiler commands)
        add_enable_option('VERBOSE', build_option('debug'))

        # Restrict parallelism
        options.append(f'MAX_JOBS={self.cfg.parallel}')

        # BLAS Interface
        if get_software_root('imkl'):
            options.append('BLAS=MKL')
            options.append('INTEL_MKL_DIR=$MKLROOT')
        elif pytorch_version >= '1.11.0' and get_software_root('FlexiBLAS'):
            options.append('BLAS=FlexiBLAS')
            options.append('WITH_BLAS=flexi')
        elif pytorch_version >= '1.9.0' and get_software_root('BLIS'):
            options.append('BLAS=BLIS')
            options.append('BLIS_HOME=' + get_software_root('BLIS'))
            options.append('USE_MKLDNN_CBLAS=ON')
        elif get_software_root('OpenBLAS'):
            # This is what PyTorch defaults to if no MKL is found.
            # Make this explicit here to avoid it finding MKL from the system
            options.append('BLAS=Eigen')
            # Still need to set a BLAS lib to use.
            # Valid choices: mkl/open/goto/acml/atlas/accelerate/veclib/generic (+blis for 1.9+)
            options.append('WITH_BLAS=open')
            # Make sure this option is actually passed to CMake
            apply_regex_substitutions(os.path.join('tools', 'setup_helpers', 'cmake.py'), [
                ("'BLAS',", "'BLAS', 'WITH_BLAS',")
            ])
        else:
            raise EasyBuildError("Did not find a supported BLAS in dependencies. Don't know which BLAS lib to use")

        available_dependency_options = EB_PyTorch.get_dependency_options_for_version(self.version)
        dependency_names = set(dep['name'] for dep in self.cfg.dependencies())
        not_used_dep_names = []
        for enable_opt, dep_name in available_dependency_options:
            if dep_name is None:
                continue
            if dep_name in dependency_names:
                options.append(enable_opt)
            else:
                not_used_dep_names.append(dep_name)
        self.log.info('Did not enable options for the following dependencies as they are not used in the EC: %s',
                      not_used_dep_names)

        # Use Infiniband by default
        # you can disable this by including 'USE_IBVERBS=0' in 'custom_opts' in the easyconfig file
        options.append('USE_IBVERBS=1')

        if get_software_root('CUDA'):
            options.append('USE_CUDA=1')
            cudnn_root = get_software_root('cuDNN')
            if cudnn_root:
                options.append('CUDNN_LIB_DIR=' + os.path.join(cudnn_root, 'lib64'))
                options.append('CUDNN_INCLUDE_DIR=' + os.path.join(cudnn_root, 'include'))

            nccl_root = get_software_root('NCCL')
            if nccl_root:
                options.append('USE_SYSTEM_NCCL=1')
                options.append('NCCL_INCLUDE_DIR=' + os.path.join(nccl_root, 'include'))

            # list of CUDA compute capabilities to use can be specifed in two ways (where (2) overrules (1)):
            # (1) in the easyconfig file, via the custom cuda_compute_capabilities;
            # (2) in the EasyBuild configuration, via --cuda-compute-capabilities configuration option;
            cuda_cc = build_option('cuda_compute_capabilities') or self.cfg['cuda_compute_capabilities']
            if not cuda_cc:
                raise EasyBuildError('List of CUDA compute capabilities must be specified, either via '
                                     'cuda_compute_capabilities easyconfig parameter or via '
                                     '--cuda-compute-capabilities')

            self.log.info('Compiling with specified list of CUDA compute capabilities: %s', ', '.join(cuda_cc))
            # This variable is also used at runtime (e.g. for tests) and if it is not set PyTorch will automatically
            # determine the compute capability of a GPU in the system and use that which may fail tests if
            # it is to new for the used nvcc
            env.setvar('TORCH_CUDA_ARCH_LIST', ';'.join(cuda_cc))
        else:
            # Disable CUDA
            options.append('USE_CUDA=0')

        if pytorch_version >= '2.0':
            add_enable_option('USE_ROCM', get_software_root('ROCm'))
        elif pytorch_version >= 'v1.10.0':
            add_enable_option('USE_MAGMA', get_software_root('magma'))

        if get_cpu_architecture() == POWER:
            # *NNPACK is not supported on Power, disable to avoid warnings
            options.extend(['USE_NNPACK=0', 'USE_QNNPACK=0', 'USE_PYTORCH_QNNPACK=0', 'USE_XNNPACK=0'])
            # Breakpad (Added in 1.10, removed in 1.12.0) doesn't support PPC
            if pytorch_version >= '1.10.0' and pytorch_version < '1.12.0':
                options.append('USE_BREAKPAD=0')
            # FBGEMM requires AVX512, so not available on PPC
            if pytorch_version >= 'v1.10.0':
                options.append('USE_FBGEMM=0')

        # Metal only supported on IOS which likely doesn't work with EB, so disabled
        options.append('USE_METAL=0')

        build_type = self.cfg.get('build_type')
        if build_type is None:
            build_type = 'Debug' if self.toolchain.options.get('debug', None) else 'Release'
        else:
            for name in ('prebuildopts', 'preinstallopts', 'custom_opts'):
                if '-DCMAKE_BUILD_TYPE=' in self.cfg[name]:
                    self.log.warning('CMAKE_BUILD_TYPE is set in %s. Ignoring build_type', name)
                    build_type = None
        if build_type:
            if pytorch_version >= '1.2.0':
                options.append('CMAKE_BUILD_TYPE=' + build_type)
            else:
                # Older versions use 2 env variables defaulting to "Release" if none are set
                build_type = build_type.lower()
                add_enable_option('DEBUG', build_type == 'debug')
                add_enable_option('REL_WITH_DEB_INFO', build_type == 'relwithdebinfo')

        unique_options = self.cfg['custom_opts']
        for option in options:
            name = option.split('=')[0] + '='  # Include the equals sign to avoid partial matches
            if not any(opt.startswith(name) for opt in unique_options):
                unique_options.append(option)

        self.cfg.update('prebuildopts', ' '.join(unique_options) + ' ')
        self.cfg.update('preinstallopts', ' '.join(unique_options) + ' ')

    def _set_cache_dir(self):
        """Set $XDG_CACHE_HOME to avoid PyTorch defaulting to $HOME"""
        cache_dir = os.path.join(self.tmpdir, '.cache')
        # The path must exist!
        mkdir(cache_dir, parents=True)
        env.setvar('XDG_CACHE_HOME', cache_dir)

    def _compare_test_results(self, old_result, xml_result, old_failed_test_names, xml_failed_test_names):
        """Compare test results parsed from stdout and XML files"""
        diffs = []

        old_suite_names = {suite.name for suite in old_result.failed_suites}
        new_suite_names = {suite.name for suite in xml_result.failed_suites}
        new_suites = new_suite_names - old_suite_names
        missing_suites = old_suite_names - new_suite_names

        diffs = []
        if new_suites:
            diffs.append(f'Found {len(new_suites)} new suites in XML files: {", ".join(sorted(new_suites))}')
        if missing_suites:
            diffs.append(f'Did not find {len(missing_suites)} suites in XML files: ' +
                         ", ".join(sorted(missing_suites)))
        if xml_result.test_cnt != old_result.test_cnt:
            diffs.append(f'Different number of tests in XML files: {xml_result.test_cnt} != {old_result.test_cnt}')
        if xml_result.error_cnt != old_result.error_cnt:
            diffs.append(f'Different number of test errors in XML files: '
                         f'{xml_result.error_cnt} != {old_result.error_cnt}')
        if xml_result.failure_cnt != old_result.failure_cnt:
            diffs.append(f'Different number of test failures in XML files: '
                         f'{xml_result.failure_cnt} != {old_result.failure_cnt}')

        def get_test_name_diff(lst_should, lst_is):
            # Handle the case where one includes the class name and the other doesn't
            return [name for name in lst_is
                    if not any(name == name2 or
                               name.endswith(f'.{name2}') or
                               name2.endswith(f'.{name}') for name2 in lst_should)]
        new_tests = get_test_name_diff(old_failed_test_names.error, xml_failed_test_names.error)
        missing_tests = get_test_name_diff(xml_failed_test_names.error, old_failed_test_names.error)
        if new_tests:
            diffs.append(f'Found {len(new_tests)} new tests with errors in XML files: {", ".join(sorted(new_tests))}')
        if missing_tests:
            diffs.append(f'Did not find {len(missing_tests)} tests with errors in XML files: ' +
                         ", ".join(sorted(missing_tests)))
        new_tests = get_test_name_diff(old_failed_test_names.fail, xml_failed_test_names.fail)
        missing_tests = get_test_name_diff(xml_failed_test_names.fail, old_failed_test_names.fail)
        if new_tests:
            diffs.append(f'Found {len(new_tests)} new failed tests in XML files: {", ".join(sorted(new_tests))}')
        if missing_tests:
            diffs.append(f'Did not find {len(missing_tests)} failed tests in XML files: ' +
                         ", ".join(sorted(missing_tests)))
        if diffs:
            self.log.warning("Found differences when parsing stdout and XML files:\n\t" + "\n\t".join(diffs))

    def test_step(self):
        """Run unit tests"""
        self._set_cache_dir()
        # Pretend to be on FB CI which disables some tests, especially those which download stuff
        env.setvar('SANDCASTLE', '1')
        # Skip this test(s) which is very flaky
        env.setvar('SKIP_TEST_BOTTLENECK', '1')
        if self.has_xml_test_reports:
            env.setvar(self.GENERATE_TEST_REPORT_VAR_NAME, '1')
        # Parse excluded_tests and flatten into space separated string
        excluded_tests = []
        for arch, tests in self.cfg['excluded_tests'].items():
            if not arch or arch == get_cpu_architecture():
                excluded_tests.extend(tests)
        # -x should not be used if there are no excluded tests
        if excluded_tests:
            excluded_tests = ['-x'] + excluded_tests
        self.cfg.template_values.update({
            'python': self.python_cmd,
            'excluded_tests': ' '.join(excluded_tests)
        })

        parsed_test_result = super().test_step(return_output_ec=True)
        if parsed_test_result is None:
            if self.cfg['runtest'] is False:
                msg = "Do not set 'runtest' to False, use --skip-test-step instead."
            else:
                msg = "Tests did not run. Make sure 'runtest' is set to a command."
            raise EasyBuildError(msg)

        tests_out, tests_ec = parsed_test_result

        failed_test_names = find_failed_test_names(tests_out)
        parsed_test_result = parse_test_log(tests_out)

        if self.has_xml_test_reports:
            test_reports_path = Path(self.start_dir) / 'test' / 'test-reports'
            try:
                xml_results = get_test_results(test_reports_path)
            except ValueError as e:
                raise EasyBuildError(f"Failed to parse test results at {test_reports_path}: {e}")
            if not xml_results:
                files = [file for file in test_reports_path.rglob('*.*') if file.is_file()]
                if files:
                    msg = f'Did not find any test result at {test_reports_path}. Files: {", ".join(files)}'
                else:
                    msg = f'Failed to find any test report files at {test_reports_path}'
                raise EasyBuildError(msg)
            missing_suites = [suite.name for suite in parsed_test_result.failed_suites
                              if suite.name not in xml_results]
            if missing_suites:
                raise EasyBuildError('Parsing the test result files missed the following failed suites: %s',
                                     ', '.join(sorted(missing_suites)))
            # Replace results as the files should be more reliable than the parsed ones
            new_result = TestResult(test_cnt=sum(suite.num_tests for suite in xml_results.values()),
                                    error_cnt=sum(suite.errors for suite in xml_results.values()),
                                    failure_cnt=sum(suite.failures for suite in xml_results.values()),
                                    failed_suites=[suite for suite in xml_results.values()
                                                   if suite.failures + suite.errors > 0],
                                    terminated_suites=parsed_test_result.terminated_suites,
                                    all_failed_suites=parsed_test_result.all_failed_suites)
            new_failed_names = FailedTestNames(
                error=list(chain.from_iterable(suite.get_errored_tests() for suite in xml_results.values())),
                fail=list(chain.from_iterable(suite.get_failed_tests() for suite in xml_results.values()))
            )
            if LooseVersion(self.version) < '2.3':
                # Show differences to results parsed from logfile. In 2.3+ the parsed values are no longer reliable
                self._compare_test_results(parsed_test_result, new_result, failed_test_names, new_failed_names)
            parsed_test_result = new_result
            failed_test_names = new_failed_names

        # Show failed subtests, if any, to aid in debugging failures
        if failed_test_names.error or failed_test_names.fail:
            msg = []
            if failed_test_names.error:
                msg.append("Found %d individual tests that exited with an error: %s"
                           % (len(failed_test_names.error), ', '.join(failed_test_names.error)))
            if failed_test_names.fail:
                msg.append("Found %d individual tests with failed assertions: %s"
                           % (len(failed_test_names.fail), ', '.join(failed_test_names.fail)))
            self.log.warning("\n".join(msg))

        # Create clear summary report
        # Use a list of messages we can later join together
        failure_msgs = ['\t%s (%s)' % (suite.name, suite.summary) for suite in parsed_test_result.failed_suites]
        # These were accounted for
        failed_test_suites = set(suite.name for suite in parsed_test_result.failed_suites)
        # Those are all that failed according to the summary output
        all_failed_test_suites = parsed_test_result.all_failed_suites
        # We should have determined all failed test suites and only those.
        # Otherwise show the mismatch and terminate later
        if failed_test_suites != all_failed_test_suites:
            failure_msgs.insert(0, 'Failed tests (suites/files):')
            # Test suites where we didn't match a specific regexp and hence likely didn't count the failures
            uncounted_test_suites = all_failed_test_suites - failed_test_suites
            if uncounted_test_suites:
                failure_msgs.append('Could not count failed tests for the following test suites/files:')
                for suite_name in sorted(uncounted_test_suites):
                    try:
                        signal = parsed_test_result.terminated_suites[suite_name]
                        reason = f'Terminated with {signal}'
                    except KeyError:
                        # Not ended with signal, might have failed due to e.g. syntax errors
                        reason = 'Undetected or did not run properly'
                    failure_msgs.append(f'\t{suite_name} ({reason})')
            # Test suites not included in the catch-all regexp but counted. Should be empty.
            unexpected_test_suites = failed_test_suites - all_failed_test_suites
            if unexpected_test_suites:
                failure_msgs.append('Counted failures of tests from the following test suites/files that are not '
                                    'contained in the summary output of PyTorch:')
                failure_msgs.extend(sorted(unexpected_test_suites))

        # Calculate total number of unsuccesful and total tests
        failed_test_cnt = parsed_test_result.failure_cnt + parsed_test_result.error_cnt
        # Only add count message if we detected any failed tests
        if failed_test_cnt > 0:
            failure_or_failures = 'failure' if parsed_test_result.failure_cnt == 1 else 'failures'
            error_or_errors = 'error' if parsed_test_result.error_cnt == 1 else 'errors'
            failure_msgs.insert(0, "%d test %s, %d test %s (out of %d):" % (
                parsed_test_result.failure_cnt, failure_or_failures,
                parsed_test_result.error_cnt, error_or_errors,
                parsed_test_result.test_cnt
            ))

        # Assemble final report
        failure_report = '\n'.join(failure_msgs)

        if failed_test_suites != all_failed_test_suites:
            # Fail because we can't be sure how many tests failed
            # so comparing to max_failed_tests cannot reasonably be done
            if failed_test_suites | set(parsed_test_result.terminated_suites) == all_failed_test_suites:
                # All failed test suites are either counted or terminated with a signal
                msg = ('Failing because these test suites were terminated which makes it impossible '
                       'to accurately count the failed tests: ')
                msg += ", ".join("%s(%s)" % name_signal
                                 for name_signal in sorted(parsed_test_result.terminated_suites.items()))
            elif len(failed_test_suites) < len(all_failed_test_suites):
                msg = ('Failing because not all failed tests could be determined. Tests failed to start, crashed '
                       'or the test accounting in the PyTorch EasyBlock needs updating!\n'
                       'Missing: ' + ', '.join(sorted(all_failed_test_suites - failed_test_suites)))
            else:
                msg = ('Failing because there were unexpected failures detected: ' +
                       ', '.join(sorted(failed_test_suites - all_failed_test_suites)))
            raise EasyBuildError(msg + '\n' +
                                 'You can check the test failures (in the log) manually and if they are harmless, '
                                 'use --ignore-test-failures to make the test step pass.\n' + failure_report)

        if failed_test_cnt > 0:
            max_failed_tests = self.cfg['max_failed_tests']

            # If no tests are supposed to fail don't print the explanation, just fail
            if max_failed_tests == 0:
                raise EasyBuildError(failure_report)
            msg = failure_report + '\n\n' + ''.join([
                "The PyTorch test suite is known to include some flaky tests, ",
                "which may fail depending on the specifics of the system or the context in which they are run.\n",
                f"For this PyTorch installation, EasyBuild allows up to {max_failed_tests} tests to fail.\n",
                "We recommend to double check that the failing tests listed above ",
                "are known to be flaky, or do not affect your intended usage of PyTorch.\n",
                "In case of doubt, reach out to the EasyBuild community (via GitHub, Slack, or mailing list).",
            ])
            # Print to console in addition to file,
            # the user should really be aware that we are accepting failing tests here...
            print_warning(msg, log=self.log)

            if failed_test_cnt > max_failed_tests:
                raise EasyBuildError("Too many failed tests (%d), maximum allowed is %d",
                                     failed_test_cnt, max_failed_tests)
        elif failure_report:
            raise EasyBuildError("Test ended with failures! Exit code: %s\n%s", tests_ec, failure_report)
        elif tests_ec:
            raise EasyBuildError("Test command had non-zero exit code (%s), but no failed tests found?!", tests_ec)

    def test_cases_step(self):
        self._set_cache_dir()
        super().test_cases_step()

    def sanity_check_step(self, *args, **kwargs):
        """Custom sanity check for PyTorch"""

        if self.cfg.get('download_dep_fail', True):
            # CMake might mistakenly download dependencies during configure
            self.log.info('Checking for downloaded submodules')
            pattern = r'^-- Downloading (\w+) to /'
            downloaded_deps = re.findall(pattern, self.install_cmd_output, re.M)

            if downloaded_deps:
                self.log.info('Found downloaded submodules: %s', ', '.join(downloaded_deps))
                fail_msg = 'found one or more downloaded dependencies: %s' % ', '.join(downloaded_deps)
                self.sanity_check_fail_msgs.append(fail_msg)

        super().sanity_check_step(*args, **kwargs)


# ###################################### Code for parsing PyTorch Test XML files ######################################
class TestState(Enum):
    """Result of a test case run"""
    SUCCESS, FAILURE, ERROR, SKIPPED = "success", "failure", "error", "skipped"


if sys.version_info >= (3, 9):
    from dataclasses import dataclass

    @dataclass
    class TestCase:
        """Instance of a test method run"""
        name: str
        state: TestState
        num_reruns: int
else:
    from collections import namedtuple
    TestCase = namedtuple('TestCase', ('name', 'state', 'num_reruns'))


class TestSuite:
    """Collection of tests in the same test file"""

    def __init__(self, name: str, errors: int, failures: int, skipped: int, test_cases: Dict[str, TestCase]):
        num_per_state = Counter(test_case.state for test_case in test_cases.values())
        if skipped != num_per_state[TestState.SKIPPED]:
            raise ValueError(f'Expected {skipped} skipped tests but found {num_per_state[TestState.SKIPPED]}')
        if failures != num_per_state[TestState.FAILURE]:
            raise ValueError(f'Expected {failures} failed tests but found {num_per_state[TestState.FAILURE]}')
        if errors != num_per_state[TestState.ERROR]:
            raise ValueError(f'Expected {errors} errored tests but found {num_per_state[TestState.ERROR]}')

        self.name = name
        self.errors = errors
        self.failures = failures
        self.skipped = skipped
        self.test_cases = test_cases

    def __getitem__(self, name: str) -> TestCase:
        """Return testcase by name"""
        return self.test_cases[name]

    def _adjust_count(self, state: TestState, val: int):
        """Adjust the relevant state count"""
        if state == TestState.FAILURE:
            self.failures += val
        elif state == TestState.SKIPPED:
            self.skipped += val
        elif state == TestState.ERROR:
            self.errors += val
        elif state != TestState.SUCCESS:
            raise ValueError(f'Invalid state {state}')

    @property
    def num_tests(self) -> int:
        """Return the total number of tests"""
        return len(self.test_cases)

    @property
    def summary(self) -> str:
        """Return a textual sumary"""
        num_passed = len(self.test_cases) - self.errors - self.failures - self.skipped
        return f'{self.failures} failed, {num_passed} passed, {self.skipped} skipped, {self.errors} errors'

    def get_tests(self) -> Iterable[TestCase]:
        """Return all test instances"""
        return self.test_cases.values()

    def add_test(self, test: TestCase):
        """Add a test instance"""
        if test.name in self.test_cases:
            raise ValueError(f"Duplicate test case '{test}' in test suite {self.name}")
        self.test_cases[test.name] = test
        self._adjust_count(test.state, 1)

    def replace_test(self, test: TestCase):
        """Replace an existing test instance"""
        existing_test = self.test_cases.pop(test.name)
        self._adjust_count(existing_test.state, -1)
        self.add_test(test)

    def get_errored_tests(self) -> List[str]:
        """Return a list of test names that exited with an error"""
        return [test.name for test in self.test_cases.values() if test.state == TestState.ERROR]

    def get_failed_tests(self) -> List[str]:
        """Return a list of failed test names"""
        return [test.name for test in self.test_cases.values() if test.state == TestState.FAILURE]


def parse_test_cases(test_suite_el: ET.Element) -> List[TestCase]:
    """Extract all test cases from the testsuite XML element"""
    test_cases: List[TestCase] = []
    for testcase in test_suite_el.iterfind("testcase"):
        classname = testcase.attrib["classname"]
        test_name = f'{classname}.{testcase.attrib["name"]}'
        failed, errored, skipped = [testcase.find(tag) is not None for tag in ("failure", "error", "skipped")]
        num_reruns = len(testcase.findall("rerun"))

        if skipped:
            if num_reruns > 0 or failed or errored:
                raise ValueError(f"Invalid state for testcase '{test_name}'")
            state = TestState.SKIPPED
        else:
            state = TestState.FAILURE if failed else TestState.ERROR if errored else TestState.SUCCESS

        test_cases.append(TestCase(test_name, state=state, num_reruns=num_reruns))
    return test_cases


def determine_suite_name(xml_file: Path, test_suite_xml: List[ET.Element]) -> str:
    """Determine main test suite name from path(s) to match against run_test.py output"""
    # Gather all file attributes from the test cases if set
    test_cases = [testcase for suite in test_suite_xml for testcase in suite.iterfind("testcase")]
    file_attribute = {testcase.attrib.get("file") for testcase in test_cases}
    file_attribute.discard(None)
    suite_name = xml_file.parent.name.replace('.', os.path.sep)  # Usually the suite name is the folder name
    if xml_file.name.startswith('TEST-'):
        # Python unittest reports have 1 file per test class:
        # test-reports/python-unittest/test_package/TEST-test_repackage.TestRepackage-20250217120914.xml
        # -> test_repackage.py ran TestRepackage
        # test-reports/dist-gloo/distributed.algorithms.test_quantization/TEST-DistQuantizationTests-20250123170925.xml
        # -> distributed/algorithms/test_quantization ran DistQuantizationTests in dist-gloo variant
        # Just do a sanity check
        if len(file_attribute) > 1:
            raise ValueError(f"Found multiple reported files in unittest report of '{xml_file}': {file_attribute}")
        reported_file = os.path.basename(file_attribute.pop())

        name_parts = xml_file.name[len('TEST-'):].rsplit('-', 1)[0].rsplit('.', 2)
        # If there is only one part it is the class -> filename is in the suite name
        if len(name_parts) == 1:
            test_file_name = os.path.basename(suite_name) + '.py'
        else:
            # Note that multiple parts are possible for sub-test files:
            # TEST-jit.test_builtins.TestBuiltins (jit/test_builtins.py)
            test_file_name = name_parts[-2] + '.py'
        if test_file_name != reported_file:
            raise ValueError(f"Unexpected file attributes in test cases of '{xml_file}'. "
                             f"Expected {test_file_name}, got {file_attribute}")
    elif suite_name == 'run_test':
        # Generic report, so try to infer from the class names which look like
        # test.distributed.pipeline.sync.test_stream.TestGetDevice or
        # test.distributed.pipeline.sync.test_pipe
        # distributed.elastic.events.lib_test.RdzvEventLibTest
        def extract_path(classname: str) -> str:
            parts = classname.split('.')
            if parts[0] == 'test':
                if not parts[-1].startswith('test_') and 'Test' in parts[-1]:
                    # last part is a real class name
                    parts.pop()
                return os.path.join(*parts[1:])
            return None
        possible_paths = {extract_path(testcase.attrib["classname"]) for testcase in test_cases}
        possible_paths.discard(None)
        if not possible_paths:
            raise ValueError("Could not infer test suite name from class names for {xml_file}.")
        # We can remove possible class names by only using the common part
        suite_name = os.path.commonpath(possible_paths)
        # Strip of common prefix to all classes, but keep the last part for uniqueness
        non_classname_prefix = os.path.dirname(suite_name).replace(os.path.sep, '.') + '.'
        for testcase in test_cases:
            classname = testcase.attrib["classname"]
            if classname.startswith(non_classname_prefix):
                testcase.attrib["classname"] = classname[len(non_classname_prefix):]
    else:
        # Pytest reports, have the name in folder and file e.g.:
        # distributed.pipeline.sync.skip.test_stash_pop/distributed.pipeline.sync.skip.test_stash_pop-052ae03efad18.xml
        # -> distributed/pipeline/sync/skip/test_stash_pop
        test_file_path = xml_file.name.rsplit('-', 1)[0].replace('.', os.path.sep)
        if test_file_path != suite_name:
            raise ValueError(f"Path from folder and filename should be equal. "
                             f"Got: '{test_file_path}' != '{suite_name}'")
    # Variant might be dist-gloo, dist-mpi or similar which is the same test code ran in different configurations!
    variant = xml_file.parent.parent.name
    if variant not in ('python-unittest', 'python-pytest'):
        suite_name = os.path.join(variant, suite_name)
    return suite_name


def parse_test_result_file(xml_file: Path) -> List[TestSuite]:
    """
    Parses the given XML file into TestSuite and TestCase objects.

    :param file_path: Path to an XML file storing test results.
    :return: A list of TestSuite objects representing the parsed structure.
    """
    root = ET.parse(xml_file).getroot()

    # Normalize root to be a list of test suite elements
    if root.tag == "testsuites":
        test_suite_xml: List[ET.Element] = root.findall("testsuite")
    elif root.tag == "testsuite":
        test_suite_xml = [root]
    else:
        raise ValueError("Root element must be <testsuites> or <testsuite>.")

    # Suite name to correctly deduplicate tests and match against run_test.py output
    suite_name = determine_suite_name(xml_file, test_suite_xml)

    test_suites: List[TestSuite] = []

    for test_suite in test_suite_xml:
        errors = int(test_suite.attrib["errors"])
        failures = int(test_suite.attrib["failures"])
        skipped = int(test_suite.attrib["skipped"])
        num_tests = int(test_suite.attrib["tests"])
        if num_tests < failures + skipped + errors:
            raise ValueError(f"Invalid test count: "
                             f"{num_tests} tests, {failures} failures, {skipped} skipped, {errors} errors")

        parsed_test_cases = parse_test_cases(test_suite)
        if not parsed_test_cases:
            # No data about the test cases or even the name of the suite, so ignore it
            if num_tests > 0:
                raise ValueError("Testsuite contains no test cases, but reports tests.")
            continue

        test_cases: Dict[str, TestCase] = {}
        for test_case in parsed_test_cases:
            if test_case.name in test_cases:
                raise ValueError(f"Duplicate test case '{test_case}' in test suite {suite_name}")
            test_cases[test_case.name] = test_case

        if len(test_cases) != num_tests:
            raise ValueError(f"Number of test cases does not match the total number of tests: "
                             f"{len(test_cases)} vs. {num_tests}")
        test_suites.append(
            TestSuite(name=suite_name, test_cases=test_cases,
                      errors=errors, failures=failures, skipped=skipped,
                      )
        )
    return test_suites


def merge_test_suites(test_suites: Iterable[TestSuite]) -> TestSuite:
    """
    Combine results for all given test suites into a single instance.
    If there is only a single instance in the input, it is returned as is.
    """
    test_suites = iter(test_suites)
    result_suite: TestSuite = next(test_suites)
    for current_suite in test_suites:
        for current_test in current_suite.get_tests():
            try:
                existing_test = result_suite[current_test.name]
            except KeyError:
                result_suite.add_test(current_test)
            else:
                if (existing_test.state == TestState.SKIPPED) != (current_test.state == TestState.SKIPPED):
                    raise ValueError(f"Mismatch in whether test was skipped or not in suite {result_suite.name}: "
                                     f"{existing_test} vs. {current_test}")
                # If test was rerun and succeeded use that
                if current_test.state == TestState.SUCCESS and existing_test.state != TestState.SUCCESS:
                    result_suite.replace_test(current_test)
    return result_suite


def get_test_results(folder: Path) -> Dict[str, TestSuite]:
    """Return a dictionary of test results contained in the folder"""
    if folder.name.startswith('test-reports'):
        folders = [folder]
    else:
        # Gather all folders containing test-reports which might be named "test-reports_1"
        # Fallback to only the folder
        folders = [cur_dir for cur_dir in folder.glob('test-reports*') if cur_dir.is_dir()] or [folder]

    files = (file for folder in folders for file in folder.rglob('*.xml'))
    test_suites = chain.from_iterable(parse_test_result_file(file) for file in files)
    get_name = attrgetter('name')
    test_suites = sorted(test_suites, key=get_name)
    return {name: merge_test_suites(suites) for name, suites in groupby(test_suites, get_name)}


def main(arg: Path):
    if arg.is_file():
        content = arg.read_text()
        m = re.search(r'cmd .*python[^ ]* run_test\.py .* exited with exit code.*output', content)
        if m:
            content = content[m.end():]
            # Heuristic for next possible text added by EasyBuild
            m = re.search(r'^== \d+-\d+-\d+ .* (pytorch\.py|EasyBuild)', content)
            if m:
                content = content[:m.start()]

        print("Failed test names: ", find_failed_test_names(content))
        print("Test result: ", parse_test_log(content))
    elif not arg.is_dir():
        msg = f'Expected a test result file or folder with XMLs to parse, got: {arg}'
        if not arg.exists():
            msg += ' which does not exist'
        raise RuntimeError(msg)
    else:
        results = get_test_results(Path(arg))
        print(f"Found {len(results)} test suites:")
        for suite in results.values():
            print(f"Suite {suite.name} {suite.num_tests}:\t{suite.summary}")
        print("Total tests:", sum(suite.num_tests for suite in results.values()))
        print("Total failures:", sum(suite.failures for suite in results.values()))
        print("Total skipped:", sum(suite.skipped for suite in results.values()))
        print("Total errors:", sum(suite.errors for suite in results.values()))
        failed_suites = [suite.name for suite in results.values() if suite.failures + suite.errors > 0]
        print(f"Failed suites ({len(failed_suites)}):\n\t" + '\n\t'.join(sorted(failed_suites)))
        failed_tests = sum((suite.get_failed_tests() for suite in results.values()), [])
        print(f"Failed tests ({len(failed_tests)}):\n\t" + '\n\t'.join(sorted(failed_tests)))
        errored_tests = sum((suite.get_errored_tests() for suite in results.values()), [])
        print(f"Errored tests ({len(errored_tests)}):\n\t" + '\n\t'.join(sorted(errored_tests)))


if __name__ == '__main__':
    main(Path(sys.argv[1]))



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for CPLEX, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""
from easybuild.tools import LooseVersion
import glob
import os
import stat

import easybuild.tools.environment as env
from easybuild.easyblocks.generic.binary import Binary
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import adjust_permissions, change_dir, mkdir
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_shell_cmd


class EB_CPLEX(Binary):
    """
    Support for installing CPLEX.
    Version 12.2 has a self-extracting binary with a Java installer
    """

    def __init__(self, *args, **kwargs):
        """Initialize CPLEX-specific variables."""
        super().__init__(*args, **kwargs)

        self.bindir = None
        self.with_python = False
        self.multi_python = 'Python' in self.cfg['multi_deps']

        # Bypass the .mod file check for GCCcore installs
        self.cfg['skip_mod_files_sanity_check'] = True

    def prepare_step(self, *args, **kwargs):
        """Prepare build environment."""
        super().prepare_step(*args, **kwargs)

        if get_software_root('Python'):
            self.with_python = True

    def install_step(self):
        """CPLEX has an interactive installer, so use Q&A"""

        tmpdir = os.path.join(self.builddir, 'tmp')
        stagedir = os.path.join(self.builddir, 'staged')
        change_dir(self.builddir)
        mkdir(tmpdir)
        mkdir(stagedir)

        env.setvar('IATEMPDIR', tmpdir)
        dst = os.path.join(self.builddir, self.src[0]['name'])

        cmd = "%s -i console" % dst

        qa = [
            (r"PRESS <ENTER> TO CONTINUE:", ''),
            (r'Press Enter to continue viewing the license agreement, or enter'
             r' "1" to accept the agreement, "2" to decline it, "3" to print it,'
             r' or "99" to go back to the previous screen\.:', '1'),
            (r'ENTER AN ABSOLUTE PATH, OR PRESS <ENTER> TO ACCEPT THE DEFAULT :', self.installdir),
            (r'IS THIS CORRECT\? \(Y/N\):', 'y'),
            (r'PRESS <ENTER> TO INSTALL:', ''),
            (r"PRESS <ENTER> TO EXIT THE INSTALLER:", ''),
            (r"CHOOSE LOCALE BY NUMBER:", ''),
            (r"Choose Instance Management Option:", ''),
            (r"No model content or proprietary data will be sent.\n1- Yes\n2- No\n"
             r"ENTER THE NUMBER OF THE DESIRED CHOICE:", '2'),
        ]
        no_qa = [r'Installing\.\.\..*\n.*------.*\n\n.*============.*\n.*$']

        run_shell_cmd(cmd, qa_patterns=qa, qa_wait_patterns=no_qa)

        # fix permissions on install dir
        perms = stat.S_IRWXU | stat.S_IXOTH | stat.S_IXGRP | stat.S_IROTH | stat.S_IRGRP
        adjust_permissions(self.installdir, perms, recursive=False, relative=False)

        # also install Python bindings if Python is included as a dependency
        if self.with_python:
            cwd = change_dir(os.path.join(self.installdir, 'python'))
            run_shell_cmd("python setup.py install --prefix=%s" % self.installdir)
            change_dir(cwd)

    def det_bindir(self):
        """Determine CPLEX bin subdirectory."""

        # avoid failing miserably under --module-only --force
        if os.path.exists(self.installdir) and os.listdir(self.installdir):
            bin_glob = 'cplex/bin/x86-64*'
            change_dir(self.installdir)
            bins = glob.glob(bin_glob)

            if len(bins) == 1:
                self.bindir = bins[0]
            elif len(bins) > 1:
                raise EasyBuildError("More than one possible path for bin found: %s", bins)
            else:
                raise EasyBuildError("No bins found using %s in %s", bin_glob, self.installdir)
        else:
            self.bindir = 'UNKNOWN'

    def make_module_extra(self):
        """Add bin dirs and lib dirs and set CPLEX_HOME and CPLEXDIR"""
        txt = super().make_module_extra()

        # avoid failing miserably under --module-only --force
        if os.path.exists(self.installdir):
            cwd = change_dir(self.installdir)
            bins = glob.glob(os.path.join('*', 'bin', 'x86-64*'))
            libs = glob.glob(os.path.join('*', 'lib', 'x86-64*', '*pic'))
            change_dir(cwd)
        else:
            bins = []
            libs = []

        txt += self.module_generator.prepend_paths('PATH', [path for path in bins])
        txt += self.module_generator.prepend_paths('LD_LIBRARY_PATH', [path for path in bins + libs])

        txt += self.module_generator.set_environment('CPLEX_HOME', os.path.join(self.installdir, 'cplex'))
        txt += self.module_generator.set_environment('CPLEXDIR', os.path.join(self.installdir, 'cplex'))

        self.log.debug("make_module_extra added %s" % txt)
        return txt

    def sanity_check_step(self):
        """Custom sanity check for CPLEX"""

        if self.bindir is None:
            self.det_bindir()

        binaries = ['cplex', 'cplexamp']
        if LooseVersion(self.version) < LooseVersion('12.8'):
            binaries.append('convert')

        custom_paths = {
            'files': [os.path.join(self.bindir, x) for x in binaries],
            'dirs': [],
        }
        custom_commands = []

        if self.with_python:
            custom_commands.append("python -c 'import cplex'")
            custom_commands.append("python -c 'import docplex'")

        super().sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



"""
EasyBuild support for building and installing OpenBLAS, implemented as an easyblock

@author: Andrew Edmondson (University of Birmingham)
@author: Alex Domingo (Vrije Universiteit Brussel)
@author: Kenneth Hoste (Ghent University)
"""
import os
import re
from easybuild.tools import LooseVersion
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.config import build_option
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.systemtools import AARCH64, POWER, get_cpu_architecture, get_shared_lib_ext
from easybuild.tools.toolchain.compiler import OPTARCH_GENERIC
import easybuild.tools.environment as env

LAPACK_TEST_TARGET = 'lapack-test'
TARGET = 'TARGET'


class EB_OpenBLAS(ConfigureMake):
    """Support for building/installing OpenBLAS."""

    @staticmethod
    def extra_options():
        """Custom easyconfig parameters for OpenBLAS easyblock."""
        extra_vars = {
            'max_failing_lapack_tests_num_errors': [0, "Maximum number of LAPACK tests failing "
                                                    "due to numerical errors", CUSTOM],
            'max_failing_lapack_tests_other_errors': [0, "Maximum number of LAPACK tests failing "
                                                      "due to non-numerical errors", CUSTOM],
            'run_lapack_tests': [False, "Run LAPACK tests during test step, "
                                        "and check whether failing tests exceeds threshold", CUSTOM],
        }

        return ConfigureMake.extra_options(extra_vars)

    def configure_step(self):
        """ set up some options - but no configure command to run"""

        default_opts = {
            'BINARY': '64',
            'CC': os.getenv('CC'),
            'FC': os.getenv('FC'),
            'MAKE_NB_JOBS': '-1',  # Disable internal parallelism to let EB choose
            'USE_OPENMP': '1',
            'USE_THREAD': '1',
        }

        if '%s=' % TARGET in self.cfg['buildopts']:
            # Add any TARGET in buildopts to default_opts, so it is passed to testopts and installopts
            for buildopt in self.cfg['buildopts'].split():
                optpair = buildopt.split('=')
                if optpair[0] == TARGET:
                    default_opts[optpair[0]] = optpair[1]
        elif LooseVersion(self.version) < LooseVersion('0.3.6') and get_cpu_architecture() == POWER:
            # There doesn't seem to be a POWER9 option yet, but POWER8 should work.
            print_warning("OpenBLAS 0.3.5 and lower have known issues on POWER systems")
            default_opts[TARGET] = 'POWER8'

        # special care must be taken when performing a generic build of OpenBLAS
        if build_option('optarch') == OPTARCH_GENERIC:
            default_opts['DYNAMIC_ARCH'] = '1'

            if get_cpu_architecture() == AARCH64:
                # when building for aarch64/generic, we also need to set TARGET=ARMV8 to make sure
                # that the driver parts of OpenBLAS are compiled generically;
                # see also https://github.com/OpenMathLib/OpenBLAS/issues/4945
                default_opts[TARGET] = 'ARMV8'

                # use -mtune=generic rather than -mcpu=generic in $CFLAGS for aarch64/generic,
                # because -mcpu=generic implies a particular -march=armv* which clashes with those used by OpenBLAS
                # when building with DYNAMIC_ARCH=1
                cflags = os.getenv('CFLAGS').replace('-mcpu=generic', '-mtune=generic')
                self.log.info("Replaced -mcpu=generic with -mtune=generic in $CFLAGS")
                env.setvar('CFLAGS', cflags)

        for key in sorted(default_opts.keys()):
            for opts_key in ['buildopts', 'testopts', 'installopts']:
                if '%s=' % key not in self.cfg[opts_key]:
                    self.cfg.update(opts_key, "%s='%s'" % (key, default_opts[key]))

        self.cfg.update('installopts', 'PREFIX=%s' % self.installdir)

    def build_step(self):
        """ Custom build step excluding the tests """

        # Equivalent to `make all` without the tests
        build_parts = []
        if LooseVersion(self.version) < LooseVersion('0.3.23'):
            build_parts += ['libs', 'netlib']
            for buildopt in self.cfg['buildopts'].split():
                if 'BUILD_RELAPACK' in buildopt and '1' in buildopt:
                    build_parts += ['re_lapack']
        # just shared is necessary and sufficient with 0.3.23 + xianyi/OpenBLAS#3983
        build_parts += ['shared']

        # Pass CFLAGS through command line to avoid redefinitions (issue xianyi/OpenBLAS#818)
        cflags = 'CFLAGS'
        if os.environ[cflags]:
            self.cfg.update('buildopts', "%s='%s'" % (cflags, os.environ[cflags]))
            del os.environ[cflags]
            self.log.info("Environment variable %s unset and passed through command line" % cflags)

        makecmd = f'make {self.parallel_flag}'

        cmd = ' '.join([self.cfg['prebuildopts'], makecmd, ' '.join(build_parts), self.cfg['buildopts']])
        run_shell_cmd(cmd)

    def check_lapack_test_results(self, test_output):
        """Check output of OpenBLAS' LAPACK test suite ('make lapack-test')."""

        # example:
        #                         -->   LAPACK TESTING SUMMARY  <--
        # SUMMARY                 nb test run     numerical error         other error
        # ================        ===========     =================       ================
        # ...
        # --> ALL PRECISIONS      4116982         4172    (0.101%)        0       (0.000%)
        test_summary_pattern = r'\s+'.join([
            r"^--> ALL PRECISIONS",
            r"(?P<test_cnt>[0-9]+)",
            r"(?P<test_fail_num_error>[0-9]+)\s+\([0-9.]+\%\)",
            r"(?P<test_fail_other_error>[0-9]+)\s+\([0-9.]+\%\)",
        ])
        regex = re.compile(test_summary_pattern, re.M)
        res = regex.search(test_output)
        if res:
            (tot_cnt, fail_cnt_num_errors, fail_cnt_other_errors) = [int(x) for x in res.groups()]
            msg = "%d LAPACK tests run - %d failed due to numerical errors - %d failed due to other errors"
            self.log.info(msg, tot_cnt, fail_cnt_num_errors, fail_cnt_other_errors)

            if fail_cnt_other_errors > self.cfg['max_failing_lapack_tests_other_errors']:
                raise EasyBuildError("Too many LAPACK tests failed due to non-numerical errors: %d (> %d)",
                                     fail_cnt_other_errors, self.cfg['max_failing_lapack_tests_other_errors'])

            if fail_cnt_num_errors > self.cfg['max_failing_lapack_tests_num_errors']:
                raise EasyBuildError("Too many LAPACK tests failed due to numerical errors: %d (> %d)",
                                     fail_cnt_num_errors, self.cfg['max_failing_lapack_tests_num_errors'])
        else:
            raise EasyBuildError("Failed to find test summary using pattern '%s' in test output: %s",
                                 test_summary_pattern, test_output)

    def test_step(self):
        """ Mandatory test step plus optional runtest"""

        run_tests = ['tests']

        if self.cfg['run_lapack_tests']:
            run_tests += [LAPACK_TEST_TARGET]

        if self.cfg['runtest']:
            run_tests += [self.cfg['runtest']]

        for runtest in run_tests:
            cmd = "%s make %s %s" % (self.cfg['pretestopts'], runtest, self.cfg['testopts'])
            res = run_shell_cmd(cmd)

            # Raise an error if any test failed
            regex = re.compile("FATAL ERROR", re.M)
            errors = regex.findall(res.output)
            if errors:
                raise EasyBuildError("Found %d fatal errors in test output!", len(errors))

            # check number of failing LAPACK tests more closely
            if runtest == LAPACK_TEST_TARGET:
                self.check_lapack_test_results(res.output)

    def sanity_check_step(self):
        """ Custom sanity check for OpenBLAS """
        custom_paths = {
            'files': ['include/cblas.h', 'include/f77blas.h', 'include/lapacke_config.h', 'include/lapacke.h',
                      'include/lapacke_mangling.h', 'include/lapacke_utils.h', 'include/openblas_config.h',
                      'lib/libopenblas.a', 'lib/libopenblas.%s' % get_shared_lib_ext()],
            'dirs': [],
        }
        super().sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing NEURON, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Maxime Boissonneault (Universite Laval, Compute Canada)
@author: Alex Domingo (Vrije Universiteit Brussel)
"""
import os
import re
import tempfile

from easybuild.easyblocks.generic.cmakemake import CMakeMake
from easybuild.easyblocks.generic.pythonpackage import det_pylibdir
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools import LooseVersion
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import build_option
from easybuild.tools.filetools import write_file
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_NEURON(CMakeMake):
    """Support for building/installing NEURON."""

    @staticmethod
    def extra_options():
        """Custom easyconfig parameters for NEURON."""
        extra_vars = {
            'paranrn': [True, "Enable support for distributed simulations.", CUSTOM],
        }
        return CMakeMake.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialisation of custom class variables for NEURON."""
        super().__init__(*args, **kwargs)

        self.python_root = None
        self.pylibdir = 'UNKNOWN'

    def prepare_step(self, *args, **kwargs):
        """Custom prepare step with python detection"""
        super().prepare_step(*args, **kwargs)

        self.python_root = get_software_root('Python')

    def configure_step(self):
        """Custom configuration procedure for NEURON."""
        # enable support for distributed simulations if desired
        if self.cfg['paranrn']:
            self.cfg.update('configopts', '-DNRN_ENABLE_MPI=ON')
        else:
            self.cfg.update('configopts', '-DNRN_ENABLE_MPI=OFF')

        # specify path to InterViews if it is available as a dependency
        interviews_root = get_software_root('InterViews')
        if interviews_root:
            self.cfg.update('configopts', f"-DIV_DIR={interviews_root} -DNRN_ENABLE_INTERVIEWS=ON")
        else:
            self.cfg.update('configopts', "-DNRN_ENABLE_INTERVIEWS=OFF")

        # optionally enable support for Python as alternative interpreter
        if self.python_root:
            python_cfgopts = " ".join([
                "-DNRN_ENABLE_PYTHON=ON",
                f"-DPYTHON_EXECUTABLE={self.python_root}/bin/python",
                "-DNRN_ENABLE_MODULE_INSTALL=ON",
                f"-DNRN_MODULE_INSTALL_OPTIONS='--prefix={self.installdir}'",
            ])
            self.cfg.update('configopts', python_cfgopts)
        else:
            self.cfg.update('configopts', "-DNRN_ENABLE_PYTHON=OFF")

        # determine Python lib dir
        self.pylibdir = det_pylibdir()

        # complete configuration with configure_method of parent
        CMakeMake.configure_step(self)

    def test_step(self):
        """Custom tests for NEURON."""
        if build_option('mpi_tests'):
            nproc = self.cfg.parallel
            try:
                hoc_file = os.path.join(self.cfg['start_dir'], 'src', 'parallel', 'test0.hoc')
                cmd = self.toolchain.mpi_cmd_for(f"bin/nrniv -mpi {hoc_file}", nproc)
                res = run_shell_cmd(cmd)
            except OSError as err:
                raise EasyBuildError("Failed to run parallel hello world: %s", err)

            valid = True
            for i in range(0, nproc):
                validate_regexp = re.compile(f"I am {i:d} of {nproc:d}")
                if not validate_regexp.search(res.output):
                    valid = False
                    break
            if res.exit_code or not valid:
                raise EasyBuildError("Validation of parallel hello world run failed.")
            self.log.info("Parallel hello world OK!")
        else:
            self.log.info("Skipping MPI testing of NEURON since MPI testing is disabled")

    def sanity_check_step(self):
        """Custom sanity check for NEURON."""
        shlib_ext = get_shared_lib_ext()

        binaries = ["mkthreadsafe", "modlunit", "neurondemo", "nocmodl", "nrngui", "nrniv", "nrnivmodl",
                    "nrnmech_makefile", "nrnpyenv.sh", "set_nrnpyenv.sh", "sortspike"]
        libs = ["nrniv", "rxdmath"]
        sanity_check_dirs = ['include', 'share/nrn']

        if self.python_root:
            sanity_check_dirs += [os.path.join("lib", "python")]
            if LooseVersion(self.version) < LooseVersion('8'):
                sanity_check_dirs += [os.path.join("lib", "python%(pyshortver)s", "site-packages")]

        # this is relevant for installations of Python packages for multiple Python versions (via multi_deps)
        # (we can not pass this via custom_paths, since then the %(pyshortver)s template value will not be resolved)
        # ensure that we only add to paths specified in the EasyConfig
        sanity_check_files = [os.path.join('bin', x) for x in binaries]
        sanity_check_files += [f'lib/lib{soname}.{shlib_ext}' for soname in libs]

        custom_paths = {
            'files': sanity_check_files,
            'dirs': sanity_check_dirs,
        }

        # run NEURON demo
        demo_dir = tempfile.mkdtemp()
        demo_inp_file = os.path.join(demo_dir, 'neurondemo.inp')
        demo_inp_cmds = '\n'.join([
            "demo(3) // load the pyramidal cell model.",
            "init()  // initialise the model",
            "t       // should be zero",
            "soma.v  // will print -65",
            "run()   // run the simulation",
            "t       // should be 5, indicating that 5ms were simulated",
            "soma.v  // will print a value other than -65, indicating that the simulation was executed",
            "quit()",
        ])
        write_file(demo_inp_file, demo_inp_cmds)

        demo_sanity_cmd = f"neurondemo < {demo_inp_file} 2>&1"
        demo_regexp_version = f'NEURON -- VERSION {self.version}'
        demo_regexp_soma = '68.134337'

        custom_commands = [
            f"{demo_sanity_cmd} | grep -c '{demo_regexp_version}'",
            f"{demo_sanity_cmd} | grep -c '{demo_regexp_soma}'",
        ]

        if self.python_root:
            custom_commands.append("python -c 'import neuron; neuron.test()'")

        super().sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_step(self, *args, **kwargs):
        """
        Custom paths of NEURON module load environment
        """
        if self.python_root:
            # location of neuron python package
            if LooseVersion(self.version) < LooseVersion('8'):
                self.module_load_environment.PYTHONPATH = [os.path.join("lib", "python*", "site-packages")]
            else:
                self.module_load_environment.PYTHONPATH = [os.path.join('lib', 'python')]

        return super().make_module_step(*args, **kwargs)

    def make_module_extra(self):
        """Define extra module entries required."""

        txt = super().make_module_extra()

        # we need to make sure the correct compiler is set in the environment,
        # since NEURON features compilation at runtime
        for var in ['CC', 'CXX', 'MPICC', 'MPICXX', 'MPICH_CC', 'MPICH_CXX']:
            val = os.getenv(var)
            if val:
                txt += self.module_generator.set_environment(var, val)
                self.log.debug(f"{var} set to {val}, adding it to module")
            else:
                self.log.debug(f"{var} not set: {os.environ.get(var, None)}")

        return txt
